import numpy as np
from numba import jit
from scipy.stats import f

from ._utils import _CheckInputs
from .base import KSampleTest, KSampleTestOutput


class MANOVA(KSampleTest):
    r"""
    Multivariate analysis of variance (MANOVA) test statistic and p-value.

    MANOVA is the current standard for multivariate `k`-sample testing.

    Notes
    -----
    The test statistic is formulated as below
    :footcite:p:`pandaNonparMANOVAIndependence2021`:

    In MANOVA, we are testing if the mean vectors of each of the `k`-samples are the
    same. Define
    :math:`\{ {x_1}_i \stackrel{iid}{\sim} F_{X_1},\ i = 1, ..., n_1 \}`,
    :math:`\{ {x_2}_j \stackrel{iid}{\sim} F_{X_2},\ j = 1, ..., n_2 \}`, ... as `k`
    groups
    of samples deriving from different a multivariate Gaussian distribution with the
    same dimensionality and same covariance matrix.
    That is, the null and alternate hypotheses are,

    .. math::

       H_0 &: \mu_1 = \mu_2 = \cdots = \mu_k, \\
       H_A &: \exists \ j \neq j' \text{ s.t. } \mu_j \neq \mu_{j'}

    Let :math:`\bar{x}_{i \cdot}` refer to the columnwise means of :math:`x_i`; that is,
    :math:`\bar{x}_{i \cdot} = (1/n_i) \sum_{j=1}^{n_i} x_{ij}`. The pooled sample
    covariance of each group, :math:`W`, is

    .. math::

       W = \sum_{i=1}^k \sum_{j=1}^{n_i} (x_{ij} - \bar{x}_{i\cdot}
       (x_{ij} - \bar{x}_{i\cdot})^T

    Next, define :math:`B` as the  sample covariance matrix of the means. If
    :math:`n = \sum_{i=1}^k n_i` and the grand mean is
    :math:`\bar{x}_{\cdot \cdot} = (1/n) \sum_{i=1}^k \sum_{j=1}^{n} x_{ij}`,

    .. math::

       B = \sum_{i=1}^k n_i (\bar{x}_{i \cdot} - \bar{x}_{\cdot \cdot})
       (\bar{x}_{i \cdot} - \bar{x}_{\cdot \cdot})^T

    Some of the most common statistics used when performing MANOVA include the Wilks'
    Lambda, the Lawley-Hotelling trace, Roy's greatest root, and
    Pillai-Bartlett trace (PBT)
    :footcite:p:`bartlettNoteTestsSignificance1939`
    :footcite:p:`raoTestsSignificanceMultivariate1948`
    (PBT was chosen to be the best of these
    as it is the most conservative
    :footcite:p:`warnePrimerMultivariateAnalysis2019`) and
    :footcite:p:`everittMonteCarloInvestigation1979`
    has shown that there are
    minimal differences in statistical power among these statistics.
    Let :math:`\lambda_1, \lambda_2, \ldots, \lambda_s` refer to the eigenvalues of
    :math:`W^{-1} B`. Here :math:`s = \min(\nu_{B}, p)` is the minimum between the
    degrees of freedom of :math:`B`, :math:`\nu_{B}` and :math:`p`. So, the PBT
    MANOVA test statistic can be written as,

    .. math::

       \mathrm{MANOVA}_{n_1, \ldots, n_k} (x, y) = \sum_{i=1}^s
       \frac{\lambda_i}{1 + \lambda_i} = \mathrm{tr} (B (B + W)^{-1})

    The p-value analytically by using the F statitic. In the case of PBT, given
    :math:`m = (|p - \nu_{B}| - 1) / 2` and :math:`r = (\nu_{W} - p - 1) / 2`, this is:

    .. math::

       F_{s(2m + s + 1), s(2r + s + 1)} = \frac{(2r + s + 1)
       \mathrm{MANOVA}_{n_1, n_2} (x, y)}{(2m + s + 1) (s -
       \mathrm{MANOVA}_{n_1, n_2} (x, y))}

    References
    ----------
    .. footbibliography::
    """

    def __init__(self):
        KSampleTest.__init__(self)

    def statistic(self, *args):
        r"""
        Calulates the MANOVA test statistic.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.

        Returns
        -------
        stat : float
            The computed MANOVA statistic.
        """
        cmean = tuple(i.mean(axis=0) for i in args)
        gmean = np.vstack(args).mean(axis=0)
        W = _compute_w(args, cmean)
        B = _compute_b(args, cmean, gmean)

        stat = np.trace(B @ np.linalg.pinv(B + W))
        self.stat = stat

        return stat

    def test(self, *args):
        r"""
        Calculates the MANOVA test statistic and p-value.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of dimensions. That is, the shapes must be `(n, p)` and
            `(m, p)`, ... where `n`, `m`, ... are the number of samples and `p` is
            the number of dimensions.

        Returns
        -------
        stat : float
            The computed MANOVA statistic.
        pvalue : float
            The computed MANOVA p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import MANOVA
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = MANOVA().test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '0.000, 1.0'
        """
        inputs = list(args)
        check_input = _CheckInputs(
            inputs=inputs,
        )
        inputs = check_input()

        N = np.sum([i.shape[0] for i in inputs])
        p = inputs[0].shape[1]
        nu_w = N - len(inputs)

        if nu_w < p:
            raise ValueError("Test cannot be run, degree of freedoms is off")

        stat = self.statistic(*inputs)
        nu_b = len(inputs) - 1
        s = np.min([p, nu_b])
        m = (np.abs(p - nu_b) - 1) / 2
        n = (nu_w - p - 1) / 2
        num = 2 * n + s + 1
        denom = 2 * m + s + 1
        pvalue = f.sf(num / denom * stat / (s - stat), s * denom, s * num)
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = None

        return KSampleTestOutput(stat, pvalue)


@jit(nopython=True, cache=True)
def _compute_w(inputs, cmean):  # pragma: no cover
    """Calculate the W matrix"""

    p = list(inputs)[0].shape[1]
    W = np.zeros((p, p))

    for i in range(len(inputs)):
        for j in range(inputs[i].shape[0]):
            W += (inputs[i][j, :] - cmean[i]) @ (inputs[i][j, :] - cmean[i]).T

    return W


@jit(nopython=True, cache=True)
def _compute_b(inputs, cmean, gmean):  # pragma: no cover
    """Calculate the B matrix"""

    p = list(inputs)[0].shape[1]
    B = np.zeros((p, p))

    for i in range(len(inputs)):
        n = inputs[i].shape[0]
        B += n * (cmean[i] - gmean) @ (cmean[i] - gmean).T

    return B



from __future__ import division

from builtins import str, range, object
from past.utils import old_div

import autograd.numpy as np
from ._utils import outer_rows
from .base import GofTest
from abc import ABC, abstractmethod
import logging

import scipy
import scipy.stats as stats
from numpy.random import default_rng


class H0Simulator(ABC):
    """
    An abstract class representing a simulator to draw samples from the
    null distribution. For some tests, these are needed to conduct the test.
    """

    def __init__(self, n_simulate, seed):
        """
        n_simulate: The number of times to simulate from the null distribution.
            Must be a positive integer.
        seed: a random seed

        From: https://github.com/wittawatj/fsic-test
        """
        assert n_simulate > 0
        self.n_simulate = n_simulate
        self.seed = seed

    @abstractmethod
    def simulate(self, gof, X):
        """
        gof: a GofTest
        dat: a Data (observed data)
        Simulate from the null distribution and return a dictionary.
        One of the item is
            sim_stats: a numpy array of stats.

        From: https://github.com/wittawatj/fsic-test
        """
        raise NotImplementedError()


class FSSDH0SimCovObs(H0Simulator):
    """
    An asymptotic null distribution simulator for FSSD.  Simulate from the
    asymptotic null distribution given by the weighted sum of chi-squares. The
    eigenvalues (weights) are computed from the covarince matrix wrt. the
    observed sample.
    This is not the correct null distribution; but has the correct asymptotic
    types-1 error at alpha.
    """

    def __init__(self, n_simulate=3000, seed=10):
        super(FSSDH0SimCovObs, self).__init__(n_simulate, seed)

    def simulate(self, gof, X, fea_tensor=None):
        """
        fea_tensor: n x d x J feature matrix
        """
        assert isinstance(gof, FSSD)
        n_simulate = self.n_simulate
        seed = self.seed
        if fea_tensor is None:
            _, fea_tensor = gof.statistic(X, return_feature_tensor=True)

        J = fea_tensor.shape[2]
        n = X.shape[0]
        # n x d*J
        Tau = fea_tensor.reshape(n, -1)
        # Make sure it is a matrix i.e, np.cov returns a scalar when Tau is
        # 1d.
        cov = np.cov(Tau.T) + np.zeros((1, 1))

        arr_nfssd, eigs = list_simulate_spectral(cov, J, n_simulate, seed=self.seed)
        return {"sim_stats": arr_nfssd}


class FSSDH0SimCovDraw(H0Simulator):
    """
    An asymptotic null distribution simulator for FSSD.  Simulate from the
    asymptotic null distribution given by the weighted sum of chi-squares. The
    eigenvalues (weights) are computed from the covarince matrix wrt. the
    sample drawn from p (the density to test against).

    - The UnnormalizedDensity p is required to implement get_datasource() method.

    From: https://github.com/wittawatj/fsic-test
    """

    def __init__(self, n_draw=2000, n_simulate=3000, seed=10):
        """
        n_draw: number of samples to draw from the UnnormalizedDensity p
        """
        super(FSSDH0SimCovDraw, self).__init__(n_simulate, seed)
        self.n_draw = n_draw

    def simulate(self, gof, fea_tensor=None):
        """
        fea_tensor: n x d x J feature matrix
        This method does not use dat.

        From: https://github.com/wittawatj/fsic-test
        """
        # p = an UnnormalizedDensity
        p = gof.p
        ds = p.get_datasource()
        if ds is None:
            raise ValueError("DataSource associated with p must be available.")
        Xdraw = ds.sample(n=self.n_draw, seed=self.seed)
        _, fea_tensor = gof.statistic(Xdraw, return_feature_tensor=True)

        X = Xdraw
        J = fea_tensor.shape[2]
        n = self.n_draw
        # n x d*J
        Tau = fea_tensor.reshape(n, -1)
        # Make sure it is a matrix i.e, np.cov returns a scalar when Tau is
        # 1d.
        cov = old_div(Tau.T.dot(Tau), n) + np.zeros((1, 1))
        n_simulate = self.n_simulate

        arr_nfssd, eigs = list_simulate_spectral(cov, J, n_simulate, seed=self.seed)
        return {"sim_stats": arr_nfssd}


class FSSD(GofTest):
    r"""
    Goodness-of-fit test using The Finite Set Stein Discrepancy statistic.
    and a set of paired test locations. The statistic is n*FSSD^2.
    The statistic can be negative because of the unbiased estimator.

    .. math::
       H_0 &: \text{ the sample follows } p \\
       H_A &: \text{ the sample does not follow } p

    :math:`p` is specified to the constructor in the form of an UnnormalizedDensity.

    Notes
    -----
    Given a known probability density :math:`p` (model) and a sample
    :math:`\{ \mathbf{x}_i \}_{i=1}^n \sim q` where :math:`q` is an unknown
    density, the GoF test tests whether or not the sample :math:`\{ \mathbf{x}_i \}_{i=1}^n`
    is distributed according to a known :math:`p`.

    The implemented test relies on a new test statistic called The Finite-Set Stein Discrepancy (FSSD) :footcite:p:`jitkrittum2017linear`.
    which is a discrepancy measure between a density and a sample. Unique features of the new goodness-of-fit test are:

        * It makes only a few mild assumptions on the distributions :math:`p` and :math:`q`. The model :math:`p`
          can take almost any form. The normalizer of :math:`p` is not assumed known. The test only assesses the goodness of
          :math:`p` through :math:`\nabla_{\mathbf{x}} \log p(\mathbf{x})` i.e., the first derivative of the log density.

        * The runtime complexity of the full test (both parameter tuning and the actual test) is
          :math:`\mathcal{O}(n)` i.e., linear in the sample size.

        * It returns a set of points (features) which indicate where :math:`p` fails to fit the data.

    The FSSD test requires that the derivative of :math:`\log p` exists.
    The test requires a technical condition called the "vanishing boundary" condition for it to be consistent.
    The condition is :math:`\lim_{\|\mathbf{x} \|\to \infty} p(\mathbf{x}) \mathbf{g}(\mathbf{x}) = \mathbf{0}` where
    :math:`\mathbf{g}` is the so called the Stein witness function which depends on the kernel and
    :math:`\nabla_{\mathbf{x}} \log p(\mathbf{x})`. For a density :math:`p` which has support everywhere e.g.,
    Gaussian, there is no problem at all. However, for a density defined on a domain
    with a boundary, one has to be careful. For example, if :math:`p` is a
    Gamma density defined on the positive orthant of :math:`\mathbb{R}`, the density itself can actually be evaluated on negative points.
    Looking at the way the Gamma density is written, there is nothing that tells the test that it cannot be evaluated on negative orthant.
    Therefore, if :math:`p` is Gamma, and the observed sample also follows :math:`p`
    (i.e., :math:`H_0` is true), the test will still reject :math:`H_0`!
    The reason is that the data do not match the left tail (in the negative region!) of the Gamma.
    It is necessary to include the fact that negative region has 0 density into the density itself.
    """

    # NULLSIM_* are constants used to choose the way to simulate from the null
    # distribution to do the test.

    # Same as NULLSIM_COVQ; but assume that sample can be drawn from p.
    # Use the drawn sample to compute the covariance.
    NULLSIM_COVP = 1

    def __init__(
        self, p, k, V, null_sim=FSSDH0SimCovObs(n_simulate=3000, seed=101), alpha=0.01
    ):
        r"""
        Parameters
        ----------
        p : an instance of UnnormalizedDensity
        k : a DifferentiableKernel object
        V : J x dx numpy array of J locations to test the difference
        null_sim : an instance of H0Simulator for simulating from the null distribution.
        alpha : significance level
        """
        super(FSSD, self).__init__(p, alpha)
        self.k = k
        self.V = V
        self.null_sim = null_sim

    def test(self, X, return_simulated_stats=False):
        r"""
        Perform the goodness-of-fit test using an FSSD test statistic
        and return values computed in a dictionary.

        Parameters
        ----------
        dat : an instance of Data

        Returns
        -------
        results : a dictionary containing alpha, p-value, test statistic,
            and null hypothesis rejection status
        """
        alpha = self.alpha
        null_sim = self.null_sim
        n_simulate = null_sim.n_simulate
        n = X.shape[0]
        J = self.V.shape[0]

        nfssd, fea_tensor = self.statistic(X, return_feature_tensor=True)
        sim_results = null_sim.simulate(self, X, fea_tensor)
        arr_nfssd = sim_results["sim_stats"]

        # approximate p-value with the permutations
        pvalue = np.mean(arr_nfssd > nfssd)

        results = {
            "alpha": self.alpha,
            "pvalue": pvalue,
            "test_stat": nfssd,
            "h0_rejected": pvalue < alpha,
            "n_simulate": n_simulate,
        }
        if return_simulated_stats:
            results["sim_stats"] = arr_nfssd
        return results

    def statistic(self, X, return_feature_tensor=False):
        r"""
        Compute the test statistic. The statistic is n*FSSD^2.

        Parameters
        ----------
        dat : an instance of Data

        Returns
        -------
        stat : the test statistic, n*FSSD^2
        """
        n = X.shape[0]

        # n x d x J
        Xi = self.feature_tensor(X)
        unscaled_mean = ustat_h1_mean_variance(Xi, return_variance=False)
        stat = n * unscaled_mean

        if return_feature_tensor:
            return stat, Xi
        else:
            return stat

    def get_H1_mean_variance(self, X):
        r"""
        Calculate the mean and variance under H1 of the test statistic (divided by
        n).

        Parameters
        ----------
        dat : an instance of Data

        Returns
        -------
        mean : the mean under the alternative hypothesis of the test statistic
        variance : the variance under the alternative hypothesis of the test statistic
        """
        Xi = self.feature_tensor(X)
        mean, variance = ustat_h1_mean_variance(Xi, return_variance=True)
        return mean, variance

    def feature_tensor(self, X):
        r"""
        Compute the feature tensor which is n x d x J.
        The feature tensor can be used to compute the statistic, and the
        covariance matrix for simulating from the null distribution.

        Parameters
        ----------
        X : n x d data numpy array

        Returns
        -------
        Xi : an n x d x J numpy array
        """
        k = self.k
        J = self.V.shape[0]
        n, d = X.shape
        # n x d matrix of gradients
        grad_logp = self.p.grad_log(X)

        K = k.eval(X, self.V)

        list_grads = np.array([np.reshape(k.gradX_y(X, v), (1, n, d)) for v in self.V])
        stack0 = np.concatenate(list_grads, axis=0)
        # a numpy array G of size n x d x J such that G[:, :, J]
        # is the derivative of k(X, V_j) with respect to X.
        dKdV = np.transpose(stack0, (1, 2, 0))

        # n x d x J tensor
        grad_logp_K = outer_rows(grad_logp, K)

        Xi = old_div((grad_logp_K + dKdV), np.sqrt(d * J))
        return Xi


def power_criterion(p, X, k, test_locs, reg=1e-2, use_unbiased=True, use_2terms=False):
    r"""
    Compute the mean and standard deviation of the statistic under H1.

    Parameters
    ----------
    use_2terms : True if the objective should include the first term in the power
        expression. This term carries the test threshold and is difficult to
        compute (depends on the optimized test locations). If True, then
        the objective will be -1/(n**0.5*sigma_H1) + n**0.5 FSSD^2/sigma_H1,
        which ignores the test threshold in the first term.

    Returns
    -------
    obj : mean/sd
    """
    n = X.shape[0]
    V = test_locs
    fssd = FSSD(p, k, V, null_sim=None)
    fea_tensor = fssd.feature_tensor(X)
    u_mean, u_variance = ustat_h1_mean_variance(
        fea_tensor, return_variance=True, use_unbiased=use_unbiased
    )

    # mean/sd criterion
    sigma_h1 = np.sqrt(u_variance + reg)
    ratio = old_div(u_mean, sigma_h1)
    if use_2terms:
        obj = old_div(-1.0, (np.sqrt(n) * sigma_h1)) + np.sqrt(n) * ratio
    else:
        obj = ratio
    return obj


def ustat_h1_mean_variance(fea_tensor, return_variance=True, use_unbiased=True):
    r"""
    Compute the mean and variance of the asymptotic normal distribution
    under H1 of the test statistic.

    Parameters
    ----------
    fea_tensor : feature tensor obtained from feature_tensor()
    return_variance : If false, avoid computing and returning the variance.
    use_unbiased : If True, use the unbiased version of the mean. Can be
        negative.

    Returns
    -------
    stat : the mean
    variance: the variance
    """
    Xi = fea_tensor
    n, d, J = Xi.shape

    assert n > 1, "Need n > 1 to compute the mean of the statistic."
    # n x d*J
    Tau = np.reshape(Xi, [n, d * J])
    if use_unbiased:
        t1 = np.sum(np.mean(Tau, 0) ** 2) * (old_div(n, float(n - 1)))
        t2 = old_div(np.sum(np.mean(Tau**2, 0)), float(n - 1))
        # stat is the mean
        stat = t1 - t2
    else:
        stat = np.sum(np.mean(Tau, 0) ** 2)

    if not return_variance:
        return stat

    # compute the variance
    # mu: d*J vector
    mu = np.mean(Tau, 0)
    variance = 4 * np.mean(np.dot(Tau, mu) ** 2) - 4 * np.sum(mu**2) ** 2
    return stat, variance


def list_simulate_spectral(cov, J, n_simulate=1000, seed=82):
    r"""
    Simulate the null distribution using the spectrums of the covariance
    matrix.  This is intended to be used to approximate the null
    distribution.

    Returns
    -------
    sim_fssds : a numpy array of simulated n*FSSD values
    eigs : eigenvalues of cov
    """
    # eigen decompose
    eigs, _ = np.linalg.eig(cov)
    eigs = np.real(eigs)
    # sort in decreasing order
    eigs = -np.sort(-eigs)
    sim_fssds = simulate_null_dist(eigs, J, n_simulate=n_simulate, seed=seed)
    return sim_fssds, eigs


def simulate_null_dist(eigs, J, n_simulate=2000, seed=7):
    r"""
    Simulate the null distribution using the spectrums of the covariance
    matrix of the U-statistic. The simulated statistic is n*FSSD^2 where
    FSSD is an unbiased estimator.

    Parameters
    ----------
    eigs : a numpy array of estimated eigenvalues of the covariance
        matrix. eigs is of length d*J, where d is the input dimension, and
    J : the number of test locations.

    Returns
    -------
    fssds : a numpy array of simulated statistics.
    """
    d = old_div(len(eigs), J)
    assert d > 0
    # draw at most d x J x block_size values at a time
    block_size = max(20, int(old_div(1000.0, (d * J))))
    fssds = np.zeros(n_simulate)
    from_ind = 0
    rng = default_rng(seed)
    while from_ind < n_simulate:
        to_draw = min(block_size, n_simulate - from_ind)
        # draw chi^2 random variables.
        chi2 = rng.standard_normal(size=(d * J, to_draw)) ** 2
        # an array of length to_draw
        sim_fssds = eigs.dot(chi2 - 1.0)
        # store
        end_ind = from_ind + to_draw
        fssds[from_ind:end_ind] = sim_fssds
        from_ind = end_ind
    return fssds


def fssd_grid_search_kernel(p, X, test_locs, list_kernel):
    r"""
    Linear search for the best kernel in the list that maximizes
    the test power criterion, fixing the test locations to V.

    Parameters
    ----------
    p : UnnormalizedDensity
    dat : a Data object
    list_kernel : list of kernel candidates

    Returns
    -------
    besti : best kernel index
    objs : array of test power criteria
    """
    V = test_locs
    n_cand = len(list_kernel)
    objs = np.zeros(n_cand)
    for i in range(n_cand):
        ki = list_kernel[i]
        objs[i] = power_criterion(p, X, ki, test_locs)
        logging.info("(%d), obj: %5.4g, k: %s" % (i, objs[i], str(ki)))

    # Widths that come early in the list
    # are preferred if test powers are equal.

    besti = objs.argmax()
    return besti, objs



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises

from ...tools import linear, power
from .. import MaxMargin


class TestMaxMarginStat:
    @pytest.mark.parametrize(
        "n, obs_stat", [(100, 0.6983550238795534), (200, 0.7026459581729388)]
    )
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 5)
        stat, pvalue = MaxMargin("Dcorr").test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)

    def test_kernstat(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        stat, pvalue = MaxMargin("Hsic").test(x, y, auto=False)

        assert_almost_equal(stat, 1.0, decimal=2)
        assert_almost_equal(pvalue, 1 / 1000, decimal=2)


class TestMaxMarginErrorWarn:
    """Tests errors and warnings."""

    def test_no_indeptest(self):
        # raises error if not indep test
        assert_raises(ValueError, MaxMargin, "abcd")


class TestMaxMarginTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            ["MaxMargin", "Dcorr"],
            sim_type="indep",
            sim="multimodal_independence",
            n=100,
            p=1,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import rot_ksamp
from .. import KSampleHHG


class TestKSampleHHG:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue", [(100, 0.515, 4.912e-10), (10, 0.777, 0.125874),],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2, noise=False)
        stat, pvalue = KSampleHHG().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=3)
        assert_almost_equal(pvalue, obs_pvalue, decimal=3)

    @pytest.mark.parametrize(
        "n", [(100)],
    )
    def test_rep(self, n):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2, noise=False)
        MPstat1 = KSampleHHG().statistic(x, y)
        MPstat2 = KSampleHHG().statistic(x, y)

        assert MPstat1 == MPstat2


class TestKSampleHHGTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        rejections = 0
        for i in range(1000):
            x, y = rot_ksamp(
                "multimodal_independence", n=100, p=1, noise=True, degree=90
            )
            stat, pvalue = KSampleHHG().test(x, y)
        if pvalue < 0.05:
            rejections += 1
        est_power = rejections / 1000

        assert_almost_equal(est_power, 0, decimal=2)



import hyppo.discrim
import hyppo.independence
import hyppo.ksample
import hyppo.time_series
import hyppo.kgof
import hyppo.tools
import hyppo.d_variate
import hyppo.conditional

__version__ = "0.5.1"



from typing import NamedTuple

import numpy as np
from numba import jit
from scipy._lib._util import MapWrapper

from ._utils import _CheckInputs
from .base import DiscriminabilityTest
from sklearn.utils import check_random_state


class DiscrimTwoSampleTestOutput(NamedTuple):
    d1: float
    d2: float
    pvalue: float


class DiscrimTwoSample(DiscriminabilityTest):
    r"""
    Two Sample Discriminability test statistic and p-value.

    Two sample test measures whether the discriminability is different for
    one dataset compared to another. More details can be described in `[1]`_.

    Let :math:`\hat D_{x_1}` denote the sample discriminability of one approach,
    and :math:`\hat D_{x_2}` denote the sample discriminability of another approach.
    Then,

    .. math::

        H_0: D_{x_1} &= D_{x_2} \\
        H_A: D_{x_1} &> D_{x_2}

    Alternatively, tests can be done for :math:`D_{x_1} < D_{x_2}` and
    :math:`D_{x_1} \neq D_{x_2}`.

    .. _[1]: https://www.biorxiv.org/content/10.1101/802629v1

    Parameters
    ----------
    is_dist : bool, default: False
        Whether `x1` and `x2` are distance matrices or not.
    remove_isolates : bool, default: True
        Whether to remove the measurements with a single instance or not.
    """

    def __init__(self, is_dist=False, remove_isolates=True):
        self.is_distance = is_dist
        self.remove_isolates = remove_isolates
        DiscriminabilityTest.__init__(self)

    def statistic(self, x, y):
        """
        Helper function that calculates the discriminability test statistic.

        Parameters
        ----------
        x, y : ndarray of float
            Input data matrices. `x` and `y` must have the same number of
            samples. That is, the shapes must be `(n, p)` and `(n, q)` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, `x` and `y` can be distance matrices,
            where the shapes must both be `(n, n)`.

        Returns
        -------
        stat : float
            The computed two sample discriminability statistic.
        """
        stat = super(DiscrimTwoSample, self).statistic(x, y)

        return stat

    def test(self, x1, x2, y, reps=1000, alt="neq", workers=-1, random_state=None):
        r"""
        Calculates the test statistic and p-value for a two sample test for
        discriminability.

        Parameters
        ----------
        x1, x2 : ndarray of float
            Input data matrices. `x1` and `x2` must have the same number of
            samples. That is, the shapes must be `(n, p)` and `(n, q)` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, `x1` and `x2` can be distance matrices,
            where the shapes must both be `(n, n)`, and ``is_dist`` must set
            to ``True`` in this case.
        y : ndarray of float
            A vector containing the sample ids for our `n` samples. Should be matched
            to the inputs such that ``y[i]`` is the corresponding label for
            ``x_1[i, :]`` and ``x_2[i, :]``.
        reps : int, optional (default: 1000)
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        alt : {"greater", "less", "neq"} (default: "neq")
            The alternative hypothesis for the test. Can test that first dataset is
            more discriminable (alt = "greater"), less discriminable (alt = "less")
            or unequal discriminability (alt = "neq").
        workers : int, optional (default: -1)
            The number of cores to parallelize the p-value computation over.
            Supply -1 to use all cores available to the Process.

        Returns
        -------
        d1 : float
            The computed discriminability score for ``x1``.
        d2 : float
            The computed discriminability score for ``x2``.
        pvalue : float
            The computed two sample test p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.discrim import DiscrimTwoSample
        >>> x1 = np.ones((100,2), dtype=float)
        >>> x2 = np.concatenate([np.zeros((50, 2)), np.ones((50, 2))], axis=0)
        >>> y = np.concatenate([np.zeros(50), np.ones(50)], axis=0)
        >>> discrim1, discrim2, pvalue = DiscrimTwoSample().test(x1, x2, y)
        >>> '%.1f, %.1f, %.2f' % (discrim1, discrim2, pvalue)
        '0.5, 1.0, 0.00'
        """

        check_input = _CheckInputs(
            [x1, x2],
            y,
            reps=reps,
            is_dist=self.is_distance,
            remove_isolates=self.remove_isolates,
        )
        x, y = check_input()
        self.x1 = np.asarray(x[0])
        self.x2 = np.asarray(x[1])
        self.y = y

        self.d1 = self.statistic(self.x1, y)
        self.d2 = self.statistic(self.x2, y)
        self.da = self.d1 - self.d2

        # use all cores to create function that parallelizes over number of reps
        random_state = np.random.randint(np.iinfo(np.int32).max, size=reps)
        mapwrapper = MapWrapper(workers)
        null_dist = np.array(list(mapwrapper(self._perm_stat, random_state)))

        self.diff_null = np.asarray(calculate_diff_null(null_dist, reps))

        if alt == "greater":
            pvalue = (self.diff_null > self.da).mean()
        elif alt == "less":
            pvalue = (self.diff_null < self.da).mean()
        elif alt == "neq":
            pvalue = (abs(self.diff_null) > abs(self.da)).mean()
        else:
            msg = "You have not entered a valid alternative."
            raise ValueError(msg)

        if pvalue == 0:
            pvalue = 1 / reps

        self.pvalue = pvalue

        return DiscrimTwoSampleTestOutput(self.d1, self.d2, self.pvalue)

    def _get_convex_comb(self, x, random_state=None):  # pragma: no cover
        """Get random convex combination of input x."""
        n, _ = x.shape

        rng = check_random_state(random_state)
        q1 = rng.choice(n, n)
        q2 = rng.choice(n, n)
        lamda = rng.uniform(size=n)

        return (lamda * (x[q1]).T + (1 - lamda) * (x[q2]).T).T

    def _perm_stat(self, index, random_state=None):  # pragma: no cover
        r"""
        Helper function that is used to calculate parallel permuted test
        statistics.

        Parameters
        ----------
        index : int
            Iterator used for parallel statistic calculation

        Returns
        -------
        perm_stat1, perm_stat2 : float
            Test statistic for each value in the null distribution.
        """
        permx1 = self._get_convex_comb(self.x1, random_state)
        permx2 = self._get_convex_comb(self.x2, random_state)

        perm_stat1 = self.statistic(permx1, self.y)
        perm_stat2 = self.statistic(permx2, self.y)

        return perm_stat1, perm_stat2


@jit(nopython=True, cache=True)
def calculate_diff_null(null_dist, reps):  # pragma: no cover
    """
    Helper function to calculate the distribution of thedifference under
    null.
    """
    diff_null = []

    for i in range(0, reps - 1):
        for j in range(i + 1, reps):
            diff_null.append(null_dist[i][0] - null_dist[j][1])
            diff_null.append(null_dist[j][1] - null_dist[i][0])

    return diff_null



import numpy as np
from numba import jit

from ..independence.dcorr import _center_distmat
from ..tools import compute_dist, perm_test
from ._utils import _CheckInputs
from .base import ConditionalIndependenceTest, ConditionalIndependenceTestOutput


class PartialDcorr(ConditionalIndependenceTest):
    r"""
    Partial Distance Covariance/Correlation (PDcov/PDcorr) test statistic and p-value.

    PDcorr is a measure of dependence between two paired random matrices
    given a third random matrix of not necessarily equal dimensions :footcite:p:`szekelyPartialDistanceCorrelation2014a`.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    use_cov : bool,
        If `True`, then the statistic will compute the covariance rather than the
        correlation.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    The statistic can be derived as follows:

    Let :math:`x`, :math:`y`, and :math:`z` be :math:`(n, p)` samples of random
    variables :math:`X`, :math:`Y` and :math:`Z`. Let :math:`D^x` be the :math:`n \times n`
    distance matrix of :math:`x`, :math:`D^y` be the :math:`n \times n` be
    the distance matrix of :math:`y`, and :math:`D^z` be the :math:`n \times n` distance
    matrix of :math:`z`. Let :math:`C^x`, :math:`C^y`, and :math:`C^z` be the unbiased centered
    distance matrices (see :class:`hyppo.independence.Dcorr` for more details). The
    partial distance covariance is defined as

    .. math::
        \mathrm{PDcov}_n (x, y; z) = \frac{1}{n(n-3)} \sum_{i\neq j}^n \left(P_{z^\perp}(x)\right)_{i,j} \left(P_{z^\perp}(y)\right)_{i,j}

    where

    .. math::
        P_{z^\perp}(x) = C^x - \frac{(C^x\cdot C^z)}{ C^z \cdot C^z) C^z

    is the orthogonal proejction of :math:`C^x` onto the subspace orthogonal to :math:`C^z`.
    The partial distance correlation is defined as

    .. math::
        \mathrm{PDcorr}_n (x, y; z) = \frac{P_{z^\perp}(x)\cdot P_{z^\perp}(y)}{|P_{z^\perp}(x)} |P_{z^\perp}(y)|}

    Equivalently, the partial distance correlation can be also defined as

    .. math::
        \mathrm{CDcorr}_n (x, y; z) =  \frac{R_{xy} - R_{xz} R_{yz}}{\sqrt{(1 - R_{xz}^2)(1 - R_{yz}^2)}}

    where :math:`R_{xy}` is the unbiased distance correlation between :math:`x` and :math:`y`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", use_cov=True, **kwargs):
        self.use_cov = use_cov
        self.compute_distance = compute_distance

        # set is_distance to true if compute_distance is None
        self.is_distance = False
        if not compute_distance:
            self.is_distance = True

        ConditionalIndependenceTest.__init__(self, **kwargs)

    def __repr__(self):
        return "PartialDcorr"

    def statistic(self, x, y, z):
        r"""
        Helper function that calculates the PDcov/PDcorr test statistic.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
            ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions. Alternatively, ``x`` and ``y`` can be
            distance matrices and ``z`` can be a similarity matrix where the
            shapes must be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed PDcov/PDcorr statistic.
        """
        check_input = _CheckInputs(x, y, z)
        x, y, z = check_input()

        if not self.is_distance:
            distx, disty, distz = compute_dist(
                x,
                y,
                z,
                metric=self.compute_distance,
            )
        else:
            distx = x
            disty = y
            distz = z

        if self.use_cov:
            stat = _pdcov(distx, disty, distz)
        else:
            stat = _pdcorr(distx, disty, distz)

        self.stat = stat
        return stat

    def test(
        self,
        x,
        y,
        z,
        reps=1000,
        workers=1,
        random_state=None,
    ):
        r"""
        Calculates the PDcov/PDcorr test statistic and p-value.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices. ``x``, ``y`` and ``z`` must have the same number
            of samples. That is, the shapes must be ``(n, p)``, ``(n, q)`` and
            ``(n, r)`` where `n` is the number of samples and `p`, `q`, and `r`
            are the number of dimensions. Alternatively, ``x`` and ``y`` can be
            distance matrices and ``z`` can be a similarity matrix where the
            shapes must be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        random_state : int, default: None
            The random_state for permutation testing to be fixed for
            reproducibility.

        Returns
        -------
        stat : float
            The computed PDcov/PDcorr statistic.
        pvalue : float
            The computed PDcov/PDcorr p-value.
        """
        check_input = _CheckInputs(x, y, z, reps=reps)
        x, y, z = check_input()

        if not self.is_distance:
            x, y, z = compute_dist(x, y, z, metric=self.compute_distance, **self.kwargs)
            self.is_distance = True

        stat, pvalue, null_dist = perm_test(
            self.statistic,
            x=x,
            y=y,
            z=z,
            reps=reps,
            workers=workers,
            is_distsim=self.is_distance,
            random_state=random_state,
        )
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = null_dist

        return ConditionalIndependenceTestOutput(stat, pvalue)


@jit(nopython=True, cache=True)
def _pdcov(distx, disty, distz):  # pragma: no cover
    """Calculate the PDcov test statistic"""
    N = distx.shape[0]
    denom = N * (N - 3)

    distx = _center_distmat(distx, bias=False)
    disty = _center_distmat(disty, bias=False)
    distz = _center_distmat(distz, bias=False)

    cov_xz = np.sum(distx * distz)
    cov_yz = np.sum(disty * distz)
    var_z = np.sum(distz * distz)

    if var_z == 0:
        stat = 0
    else:
        proj_xz = distx - (cov_xz / var_z) * distz
        proj_yz = disty - (cov_yz / var_z) * distz
        stat = np.sum(proj_xz * proj_yz) / denom

    return stat


@jit(nopython=True, cache=True)
def _pdcorr(distx, disty, distz):  # pragma: no cover
    """Calculate the PDcorr test statistic"""

    distx = _center_distmat(distx, bias=False)
    disty = _center_distmat(disty, bias=False)
    distz = _center_distmat(distz, bias=False)

    cov_xy = np.sum(distx * disty)
    cov_xz = np.sum(distx * distz)
    cov_yz = np.sum(disty * distz)
    var_x = np.sum(distx * distx)
    var_y = np.sum(disty * disty)
    var_z = np.sum(distz * distz)

    if var_x <= 0 or var_y <= 0:
        rxy = 0
    else:
        rxy = cov_xy / np.real(np.sqrt(var_x * var_y))

    if var_x <= 0 or var_z <= 0:
        rxz = 0
    else:
        rxz = cov_xz / np.real(np.sqrt(var_x * var_z))

    if var_y <= 0 or var_z <= 0:
        ryz = 0
    else:
        ryz = cov_yz / np.real(np.sqrt(var_y * var_z))

    stat = (rxy - rxz * ryz) / np.sqrt((1 - rxz**2) * (1 - ryz**2))

    return stat



import numpy as np

from ._utils import _CheckInputs
from .base import IndependenceTest


class RV(IndependenceTest):
    r"""
    Rank Value (RV) test statistic and p-value.

    RV is the multivariate generalization of the squared Pearson correlation
    coefficient
    :footcite:p:`robertUnifyingToolLinear1976`.
    The RV coefficient can be thought to be closely
    related to principal component analysis (PCA), canonical correlation
    analysis (CCA), multivariate regression, and statistical classification
    :footcite:p:`robertUnifyingToolLinear1976`.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`robertUnifyingToolLinear1976`
    :footcite:p:`escoufierTraitementVariablesVectorielles1973`:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. We can center :math:`x` and :math:`y` and then
    calculate the sample covariance matrix :math:`\hat{\Sigma}_{xy} = x^T y`
    and the variance matrices for :math:`x` and :math:`y` are defined
    similarly. Then, the RV test statistic is found by calculating

    .. math::

        \mathrm{RV}_n (x, y) =
            \frac{\mathrm{tr} \left( \hat{\Sigma}_{xy}
                                     \hat{\Sigma}_{yx} \right)}
            {\mathrm{tr} \left( \hat{\Sigma}_{xx}^2 \right)
             \mathrm{tr} \left( \hat{\Sigma}_{yy}^2 \right)}

    where :math:`\mathrm{tr} (\cdot)` is the trace operator.

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self):
        IndependenceTest.__init__(self)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the RV test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples and dimensions. That is, the shapes must be ``(n, p)`` where
            `n` is the number of samples and `p` is the number of dimensions.

        Returns
        -------
        stat : float
            The computed RV statistic.
        """
        centx = x - np.mean(x, axis=0)
        centy = y - np.mean(y, axis=0)

        # calculate covariance and variances for inputs
        covar = centx.T @ centy
        varx = centx.T @ centx
        vary = centy.T @ centy

        covar = np.trace(covar @ covar.T)
        stat = np.divide(
            covar, np.sqrt(np.trace(varx @ varx)) * np.sqrt(np.trace(vary @ vary))
        )
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the RV test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples and dimensions. That is, the shapes must be ``(n, p)`` where
            `n` is the number of samples and `p` is the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed RV statistic.
        pvalue : float
            The computed RV p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import RV
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = RV().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '1.0, 0.00'
        """
        check_input = _CheckInputs(x, y, reps=reps)
        x, y = check_input()

        return super(RV, self).test(
            x, y, reps, workers, is_distsim=False, random_state=random_state
        )



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

from .. import DiscrimTwoSample


# TODO: timeout CircleCI, make discrim faster
@pytest.mark.skip(reason="no way of currently testing this")
class TestTwoSample:
    def test_greater(self):
        # test whether discriminability for x1 is greater than it is for x2
        x1 = np.ones((100, 2), dtype=float)
        x2 = np.concatenate((np.zeros((50, 2)), np.ones((50, 2))), axis=0)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        np.random.seed(123456789)
        obs_d1 = 0.5
        obs_d2 = 1.0
        d1, d2, _ = DiscrimTwoSample().test(x1, x2, y, alt="greater", reps=3)

        assert_almost_equal(d1, obs_d1, decimal=2)
        assert_almost_equal(d2, obs_d2, decimal=2)

    def test_less(self):
        # test whether discriminability for x1 is less than it is for x2
        x1 = np.ones((100, 2), dtype=float)
        x2 = np.concatenate((np.zeros((50, 2)), np.ones((50, 2))), axis=0)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        np.random.seed(123456789)
        obs_d1 = 0.5
        obs_d2 = 1.0
        d1, d2, _ = DiscrimTwoSample().test(x1, x2, y, alt="less", reps=3)

        assert_almost_equal(d1, obs_d1, decimal=2)
        assert_almost_equal(d2, obs_d2, decimal=2)

    def test_neq(self):
        # test whether discriminability for x1 is not equal compared to
        # discriminability for x2
        x1 = np.ones((100, 2), dtype=float)
        x2 = np.concatenate((np.zeros((50, 2)), np.ones((50, 2))), axis=0)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        np.random.seed(123456789)
        obs_d1 = 0.5
        obs_d2 = 1.0
        d1, d2, _ = DiscrimTwoSample().test(x1, x2, y, alt="neq", reps=3)

        assert_almost_equal(d1, obs_d1, decimal=2)
        assert_almost_equal(d2, obs_d2, decimal=2)


class TestDiscrErrorWarn:
    """Tests errors and warnings."""

    def test_no_indeptest(self):
        # raises error if not indep test
        x1 = np.arange(20)
        x2 = np.arange(3, 23)
        y = np.arange(5, 25)
        assert_raises(ValueError, DiscrimTwoSample().test, x1, x2, y, alt="abcd")



from __future__ import division

from abc import ABC, abstractmethod


class GofTest(ABC):
    r"""
    A base class for a discriminability test.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean" or "gaussian"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,
            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.
        Alternatively, this function computes the kernel similarity among the
        samples within each data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,
            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]
        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
    """

    def __init__(self, p, alpha):
        """
        p : an UnnormalizedDensity object
        alpha : float, significance level of the test
        """
        self.p = p
        self.alpha = alpha

    @abstractmethod
    def test(self, X):
        """
        Perform the goodness-of-fit test and return values
        computed in a dictionary.

        Parameters
        ----------
        dat : an instance of Data (observed data)

        Returns
        -------
        {
            alpha: 0.01,
            pvalue: 0.0002,
            test_stat: 2.3,
            h0_rejected: True,
            time_secs: ...
        }
        """
        raise NotImplementedError()

    @abstractmethod
    def statistic(self, X):
        r"""
        Calculates the goodness-of-fit test statistic.

        Parameters
        ----------
        dat : an instance of Data (observed data)
            Input data matrices.
        """
        raise NotImplementedError()



"""
Module containing kernel related classes

TODO: Replace kernel evaluation with compute_kern found in hyppo.common.tools by reducing
dependencies on autograd.numpy
"""
from __future__ import division

from builtins import str
from past.utils import old_div

from abc import ABC, abstractmethod
import autograd
import autograd.numpy as np

from ..tools import compute_kern


class Kernel(ABC):
    """Abstract class for kernels. Inputs to all methods are numpy arrays."""

    @abstractmethod
    def eval(self, X, Y):
        """
        Evaluate the kernel on data X and Y
        X: nx x d where each row represents one point
        Y: ny x d
        return nx x ny Gram matrix
        """
        pass

    @abstractmethod
    def pair_eval(self, X, Y):
        """Evaluate k(x1, y1), k(x2, y2), ...
        X: n x d where each row represents one point
        Y: n x d
        return a 1d numpy array of length n.
        """
        pass


class DifferentiableKernel(ABC):
    def gradX_y(self, X, y):
        """
        Compute the gradient with respect to X (the first argument of the
        kernel). Base class provides a default autograd implementation for convenience.
        Subclasses should override if this does not work.
        X: nx x d numpy array.
        y: numpy array of length d.
        Return a numpy array G of size nx x d, the derivative of k(X, y) with
        respect to X.
        """
        yrow = np.reshape(y, (1, -1))
        f = lambda X: self.eval(X, yrow)
        g = autograd.elementwise_grad(f)
        G = g(X)
        assert G.shape[0] == X.shape[0]
        assert G.shape[1] == X.shape[1]
        return G


class LinearKSTKernel(ABC):
    """
    Interface specifiying methods a kernel has to implement to be used with
    the linear-time version of Kernelized Stein discrepancy test of
    Liu et al., 2016 (ICML 2016).
    """

    @abstractmethod
    def pair_gradX_Y(self, X, Y):
        """
        Compute the gradient with respect to X in k(X, Y), evaluated at the
        specified X and Y.
        X: n x d
        Y: n x d
        Return a numpy array of size n x d
        """
        raise NotImplementedError()

    @abstractmethod
    def pair_gradY_X(self, X, Y):
        """
        Compute the gradient with respect to Y in k(X, Y), evaluated at the
        specified X and Y.
        X: n x d
        Y: n x d
        Return a numpy array of size n x d
        """
        raise NotImplementedError()

    @abstractmethod
    def pair_gradXY_sum(self, X, Y):
        """
        Compute \sum_{i=1}^d \frac{\partial^2 k(X, Y)}{\partial x_i \partial y_i}
        evaluated at each x_i in X, and y_i in Y.
        X: n x d numpy array.
        Y: n x d numpy array.
        Return a one-dimensional length-n numpy array of the derivatives.
        """
        raise NotImplementedError()


class KSTKernel(ABC):
    """
    Interface specifiying methods a kernel has to implement to be used with
    the Kernelized Stein discrepancy test of Chwialkowski et al., 2016 and
    Liu et al., 2016 (ICML 2016 papers) See goftest.KernelSteinTest.
    """

    @abstractmethod
    def gradX_Y(self, X, Y, dim):
        """
        Compute the gradient with respect to the dimension dim of X in k(X, Y).
        X: nx x d
        Y: ny x d
        Return a numpy array of size nx x ny.
        """
        raise NotImplementedError()

    @abstractmethod
    def gradY_X(self, X, Y, dim):
        """
        Compute the gradient with respect to the dimension dim of Y in k(X, Y).
        X: nx x d
        Y: ny x d
        Return a numpy array of size nx x ny.
        """
        raise NotImplementedError()

    @abstractmethod
    def gradXY_sum(self, X, Y):
        """
        Compute \sum_{i=1}^d \frac{\partial^2 k(x, Y)}{\partial x_i \partial y_i}
        evaluated at each x_i in X, and y_i in Y.
        X: nx x d numpy array.
        Y: ny x d numpy array.
        Return a nx x ny numpy array of the derivatives.
        """
        raise NotImplementedError()


class KGauss(DifferentiableKernel, KSTKernel, LinearKSTKernel):
    """
    The standard isotropic Gaussian kernel.
    Parameterization is the same as in the density of the standard normal
    distribution. sigma2 is analogous to the variance.
    """

    def __init__(self, sigma2):
        self.sigma2 = sigma2

    def eval(self, X, Y):
        """
        Evaluate the Gaussian kernel on the two 2d numpy arrays.
        Parameters
        ----------
        X : n1 x d numpy array
        Y : n2 x d numpy array
        Return
        ------
        K : a n1 x n2 Gram matrix.
        """
        sumx2 = np.reshape(np.sum(X**2, 1), (-1, 1))
        sumy2 = np.reshape(np.sum(Y**2, 1), (1, -1))
        D2 = sumx2 - 2 * np.dot(X, Y.T) + sumy2
        K = np.exp(old_div(-D2, (2.0 * self.sigma2)))
        return K

    def gradX_Y(self, X, Y, dim):
        """
        Compute the gradient with respect to the dimension dim of X in k(X, Y).
        X: nx x d
        Y: ny x d
        Return a numpy array of size nx x ny.
        """
        sigma2 = self.sigma2
        K = self.eval(X, Y)
        Diff = X[:, [dim]] - Y[:, [dim]].T
        G = -K * Diff / sigma2
        return G

    def pair_gradX_Y(self, X, Y):
        """
        Compute the gradient with respect to X in k(X, Y), evaluated at the
        specified X and Y.
        X: n x d
        Y: n x d
        Return a numpy array of size n x d
        """
        sigma2 = self.sigma2
        Kvec = self.pair_eval(X, Y)
        # n x d
        Diff = X - Y
        G = -Kvec[:, np.newaxis] * Diff / sigma2
        return G

    def gradY_X(self, X, Y, dim):
        """
        Compute the gradient with respect to the dimension dim of Y in k(X, Y).
        X: nx x d
        Y: ny x d
        Return a numpy array of size nx x ny.
        """
        return -self.gradX_Y(X, Y, dim)

    def pair_gradY_X(self, X, Y):
        """
        Compute the gradient with respect to Y in k(X, Y), evaluated at the
        specified X and Y.
        X: n x d
        Y: n x d
        Return a numpy array of size n x d
        """
        return -self.pair_gradX_Y(X, Y)

    def gradXY_sum(self, X, Y):
        r"""
        Compute \sum_{i=1}^d \frac{\partial^2 k(X, Y)}{\partial x_i \partial y_i}
        evaluated at each x_i in X, and y_i in Y.
        X: nx x d numpy array.
        Y: ny x d numpy array.
        Return a nx x ny numpy array of the derivatives.
        """
        (n1, d1) = X.shape
        (n2, d2) = Y.shape
        d = d1
        sigma2 = self.sigma2
        D2 = np.sum(X**2, 1)[:, np.newaxis] - 2 * np.dot(X, Y.T) + np.sum(Y**2, 1)
        K = np.exp(old_div(-D2, (2.0 * sigma2)))
        G = K / sigma2 * (d - old_div(D2, sigma2))
        return G

    def pair_gradXY_sum(self, X, Y):
        """
        Compute \sum_{i=1}^d \frac{\partial^2 k(X, Y)}{\partial x_i \partial y_i}
        evaluated at each x_i in X, and y_i in Y.
        X: n x d numpy array.
        Y: n x d numpy array.
        Return a one-dimensional length-n numpy array of the derivatives.
        """
        d = X.shape[1]
        sigma2 = self.sigma2
        D2 = np.sum((X - Y) ** 2, 1)
        Kvec = np.exp(old_div(-D2, (2.0 * self.sigma2)))
        G = Kvec / sigma2 * (d - old_div(D2, sigma2))
        return G

    def pair_eval(self, X, Y):
        """
        Evaluate k(x1, y1), k(x2, y2), ...
        Parameters
        ----------
        X, Y : n x d numpy array
        Return
        -------
        a numpy array with length n
        """
        D2 = np.sum((X - Y) ** 2, 1)
        Kvec = np.exp(old_div(-D2, (2.0 * self.sigma2)))
        return Kvec



from ._utils import sim_matrix
from .cca import CCA
from .dcorr import Dcorr
from .hhg import HHG
from .hsic import Hsic
from .kmerf import KMERF
from .max_margin import MaxMargin
from .mgc import MGC
from .rv import RV

__all__ = [s for s in dir()]  # add imported tests to __all__


INDEP_TESTS = {
    "rv": RV,
    "cca": CCA,
    "hhg": HHG,
    "hsic": Hsic,
    "dcorr": Dcorr,
    "mgc": MGC,
    "kmerf": KMERF,
    "maxmargin": MaxMargin,
}



import numpy as np

from sklearn.utils import check_array
from ..tools import check_reps, contains_nan


class _CheckInputs:
    """Checks inputs for d_variate independence tests"""

    def __init__(self, *args, reps=None):
        self.args = args
        self.reps = reps

    def __call__(self):
        for mat in self.args:
            check_array(mat)
            contains_nan(mat)
            self._check_dim(mat)
            self._convert_float64(mat)
            self._check_min_samples(mat)
            self._check_variance(mat)

        if self.reps:
            check_reps(self.reps)

        return self.args

    def _check_dim(self, mat):
        """Converts input data matrices to proper dimensions"""
        if mat.ndim == 1:
            mat = mat[:, np.newaxis]
        elif mat.ndim != 2:
            raise ValueError(
                "Expected a 2-D array, found shape " "{}".format(mat.shape)
            )

        self._check_nd_indeptest(mat)

        return mat

    def _convert_float64(self, mat):
        """Converts input data matrices to np.float 64 (if not already done)"""
        mat = np.asarray(mat).astype(np.float64)

        return mat

    def _check_nd_indeptest(self, mat):
        """Check if number of samples is the same"""
        samples = []
        n, _ = mat.shape
        samples.append(n)
        if any(sample != samples[0] for sample in samples):
            raise ValueError(
                "Shape mismatch, all input data matrices must have shape "
                "[n, p] and [n, q]."
            )

    def _check_min_samples(self, mat):
        """Check if the number of samples is at least 3"""
        samples = []
        n = mat.shape[0]
        samples.append(n)
        if any(sample <= 2 for sample in samples):
            raise ValueError("Number of samples is too low. Must be at least 3.")

    def _check_variance(self, mat):
        if np.var(mat) == 0:
            raise ValueError("Test cannot be run, one of the inputs has 0 variance")



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

from ...tools import linear, logarithmic, power
from .. import KCI


class TestKCI:
    @pytest.mark.parametrize("n", [100])
    @pytest.mark.parametrize("obs_stat", [544.691])
    @pytest.mark.parametrize("obs_pvalue", [0])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat1, pvalue1 = KCI().test(x, y)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(pvalue1, obs_pvalue, decimal=2)

    @pytest.mark.parametrize("n", [100])
    @pytest.mark.parametrize("obs_stat", [88.526])
    @pytest.mark.parametrize("obs_pvalue", [.00027])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = logarithmic(n, 1)
        stat1, pvalue1 = KCI().test(x, y)

        assert_almost_equal(stat1, obs_stat, decimal=2)
        assert_almost_equal(pvalue1, obs_pvalue, decimal=2)

    @pytest.mark.parametrize("n", [100, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        stat1, pvalue1 = KCI().test(x, y)
        stat2, pvalue2 = KCI().test(x, y)

        assert stat1 == stat2
        assert pvalue1 == pvalue2



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises

from ...tools import power, rot_ksamp
from .. import MANOVA


class TestManova:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [(1000, 0.005062841807278008, 1.0), (100, 8.24e-5, 0.9762956529114515)],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2, noise=False)
        stat, pvalue = MANOVA().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=1)


class TestManovaErrorWarn:
    """Tests errors and warnings derived from MGC."""

    def test_no_indeptest(self):
        # raises error if not indep test
        x = np.arange(100).reshape(5, 20)
        y = np.arange(50, 150).reshape(5, 20)
        assert_raises(ValueError, MANOVA().test, x, y)


class TestManovaTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "MANOVA",
            sim_type="gauss",
            sim="multimodal_independence",
            case=1,
            n=100,
            alpha=0.05,
        )

        assert est_power <= 0.05



import warnings

import numpy as np

from ..independence import MGC
from ..tools import (
    check_ndarray_xy,
    check_reps,
    compute_dist,
    contains_nan,
    convert_xy_float64,
)


class _CheckInputs:
    def __init__(self, x, y, max_lag=None, reps=None):
        self.x = x
        self.y = y
        self.max_lag = max_lag
        self.reps = reps

    def __call__(self):
        check_ndarray_xy(self.x, self.y)
        contains_nan(self.x)
        contains_nan(self.y)
        self.x, self.y = self.check_dim_xy()
        self.x, self.y = convert_xy_float64(self.x, self.y)
        self._check_min_samples()

        if self.reps:
            check_reps(self.reps)

        self.max_lag = check_max_lag(self.max_lag, self.x.shape[0])

        return self.x, self.y, self.max_lag

    def check_dim_xy(self):
        # check if x and y are ndarrays
        # convert arrays of type (n,) to (n, 1)
        if self.x.ndim == 1:
            self.x = self.x[:, np.newaxis]
        elif self.x.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `x`, found shape " "{}".format(self.x.shape)
            )
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
        elif self.y.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `y`, found shape " "{}".format(self.y.shape)
            )

        self._check_nd_indeptest()

        return self.x, self.y

    def _check_nd_indeptest(self):
        nx = self.x.shape[0]
        ny = self.y.shape[0]

        if nx != ny:
            raise ValueError(
                "Shape mismatch, x and y must have shape [n, p] and [n, q]."
            )

    def _check_min_samples(self):
        nx = self.x.shape[0]
        ny = self.y.shape[0]

        if nx <= 3 or ny <= 3:
            raise ValueError("Number of samples is too low")


def check_max_lag(max_lag, n):
    if max_lag is None:
        max_lag = int(np.ceil(np.log(n)))
    elif not np.issubdtype(type(max_lag), np.integer):
        raise ValueError("max_lag must be an integer.")
    elif max_lag < 0:
        raise ValueError("max_lag must be >= 0.")

    return max_lag


def compute_stat(x, y, indep_test, compute_distance, max_lag, **kwargs):
    """Compute time series test statistic"""
    # calculate distance matrices
    distx, disty = compute_dist(x, y, metric=compute_distance, **kwargs)

    # calculate dep_lag when max_lag is 0
    dep_lag = []
    indep_test = indep_test(compute_distance=compute_distance, **kwargs)
    indep_test_stat = indep_test.statistic(x, y)
    dep_lag.append(indep_test_stat)

    # loop over time points and find max test statistic
    n = distx.shape[0]
    for j in range(1, max_lag + 1):
        slice_distx = distx[j:n, j:n]
        slice_disty = disty[0 : (n - j), 0 : (n - j)]
        stat = indep_test.statistic(slice_distx, slice_disty)
        dep_lag.append((n - j) * stat / n)

    # calculate optimal lag and test statistic
    opt_lag = np.argmax(dep_lag)
    stat = np.sum(dep_lag)

    return stat, opt_lag


def compute_scale_at_lag(x, y, opt_lag, compute_distance, **kwargs):
    """Run the mgc test at the optimal scale (by shifting the series)."""
    n = x.shape[0]
    if not compute_distance:
        compute_distance = "precomputed"
    distx, disty = compute_dist(x, y, metric=compute_distance, **kwargs)

    slice_distx = distx[opt_lag:n, opt_lag:n]
    slice_disty = disty[0 : (n - opt_lag), 0 : (n - opt_lag)]

    mgc = MGC()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        opt_scale = mgc.test(slice_distx, slice_disty, reps=0)[2]["opt_scale"]

    return opt_scale



from abc import ABC, abstractmethod
from typing import NamedTuple

from ..tools import multi_perm_test


class DVariateTestOutput(NamedTuple):
    stat: float
    pvalue: float


class DVariateTest(ABC):
    r"""
    A base class for a :math:`d`-variate independence test.

    Parameters
    ----------
    compute_kernel : str, callable, or None, default: "gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``args`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguments to send to your custom function.
    **kwargs
        Arbitrary keyword arguments for ``multi_compute_kern``.
    """

    def __init__(self, compute_kernel=None, **kwargs):
        # set statistic and p-value
        self.stat = None
        self.pvalue = None
        self.compute_kernel = compute_kernel
        self.kwargs = kwargs

        super().__init__()

    @abstractmethod
    def statistic(self, *args):
        r"""
        Calculates the :math:`d`-variate independence test statistic.

        Parameters
        ----------
        *args: ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
            etc., where `n` is the number of samples and `p` and `q` are the
            number of dimensions.
        """

    @abstractmethod
    def test(self, *args, reps=1000, workers=1):
        r"""
        Calculates the d_variate independence test statistic and p-value.

        Parameters
        ----------
        *args : ndarray of float
            Variable length input data matrices. All inputs must have the same
            number of samples. That is, the shapes must be ``(n, p)``, ``(n, q)``,
            etc., where `n` is the number of samples and `p` and `q` are the
            number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed :math:`d`-variate independence test statistic.
        pvalue : float
            The computed :math:`d`-variate independence p-value.
        """
        self.args = args

        stat, pvalue, null_dist = multi_perm_test(
            self.statistic,
            *args,
            reps=reps,
            workers=workers,
        )
        self.stat = stat
        self.pvalue = pvalue
        self.null_dist = null_dist

        return DVariateTestOutput(stat, pvalue)



import numpy as np

from ..tools import check_ndarray_xyz, check_reps, contains_nan, convert_xyz_float64


class _CheckInputs:
    """Checks inputs for all independence tests"""

    def __init__(self, x, y, z, reps=None, max_dims=None, ignore_z_var=False):
        self.x = x
        self.y = y
        self.z = z
        self.reps = reps
        self.max_dims = max_dims
        self.ignore_z_var = ignore_z_var # to allow for constant z input

    def __call__(self):
        check_ndarray_xyz(self.x, self.y, self.z)
        contains_nan(self.x)
        contains_nan(self.y)
        contains_nan(self.z)
        self.x, self.y, self.z = self.check_dim_xyz(max_dims=self.max_dims)
        self.x, self.y, self.z = convert_xyz_float64(self.x, self.y, self.z)
        self._check_min_samples()
        self._check_variance()

        if self.reps:
            check_reps(self.reps)

        return self.x, self.y, self.z

    def check_dim_xyz(self, max_dims):
        """Check and convert x and y to proper dimensions"""
        if self.x.ndim == 1:
            self.x = self.x[:, np.newaxis]
        elif self.x.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `x`, found shape " "{}".format(self.x.shape)
            )
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
        elif self.y.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `y`, found shape " "{}".format(self.y.shape)
            )
        if self.z.ndim == 1:
            self.z = self.z[:, np.newaxis]
        elif self.z.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `z`, found shape " "{}".format(self.z.shape)
            )

        if max_dims is not None:
            _, dx = self.x.shape
            _, dy = self.y.shape
            _, dz = self.z.shape

            if np.any(np.array([dx, dy, dz]) > max_dims):
                raise ValueError(
                    f"x, y, z must have be univariate and have shape [n,{max_dims}]"
                )

        self._check_nd_indeptest()

        return self.x, self.y, self.z

    def _check_nd_indeptest(self):
        """Check if number of samples is the same"""
        nx, _ = self.x.shape
        ny, _ = self.y.shape
        nz, _ = self.z.shape
        if not np.all(np.array([nx, ny, nz]) == nx):
            raise ValueError(
                "Shape mismatch, x, y and z must have shape "
                + "[n, p], [n, q] and [n, r]."
            )

    def _check_min_samples(self):
        """Check if the number of samples is at least 3"""
        nx = self.x.shape[0]
        ny = self.y.shape[0]
        nz = self.z.shape[0]

        if nx <= 3 or ny <= 3 or nz <= 3:
            raise ValueError("Number of samples is too low")

    def _check_variance(self):
        if np.var(self.x) == 0:
        # or np.var(self.y) == 0 or np.var(self.z) == 0:
            raise ValueError("Test cannot be run. Input array x has 0 variance.")
        if np.var(self.y) == 0:
            raise ValueError("Test cannot be run. Input array y has 0 variance")
        if not self.ignore_z_var:
            if np.var(self.z) == 0:
                raise ValueError("Test cannot be run. Input array z has 0 variance")



import numpy as np

from ..tools import chi2_approx
from ._utils import _CheckInputs
from .base import IndependenceTest, IndependenceTestOutput
from .cca import CCA
from .dcorr import Dcorr
from .hhg import HHG
from .hsic import Hsic
from .kmerf import KMERF
from .mgc import MGC
from .rv import RV

INDEP_NOT_MAXMARGIN = {
    "rv": RV,
    "cca": CCA,
    "hhg": HHG,
    "hsic": Hsic,
    "dcorr": Dcorr,
    "mgc": MGC,
    "kmerf": KMERF,
}


class MaxMargin(IndependenceTest):
    r"""
    Maximal Margin test statistic and p-value.

    This test loops over each of the dimensions of the inputs :math:`x` and :math:`y`
    and computes the desired independence test statistic. Then, the maximial test
    statistic is chosen :footcite:p:`shenHighDimensionalIndependenceTesting2020`.

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    Parameters
    ----------
    indep_test : "CCA", "Dcorr", "HHG", "RV", "Hsic", "MGC", "KMERF"
        A string corresponding to the desired independence test from
        :mod:`hyppo.independence`. This is not case sensitive.
    compute_distkern : str, callable, or None, default: "euclidean" or "gaussian"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Alternatively, this function computes the kernel similarity among the
        samples within each data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics (for
        ``indep_test="Dcorr"`` and ``indep_test="Hsic"``).
    **kwargs
        Arbitrary keyword arguments for ``compute_distkern``.

    Notes
    -----
    .. note::
       This algorithm is currently under review at a peer-review journal.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, indep_test, compute_distkern="euclidean", bias=False, **kwargs):
        indep_test = indep_test.lower()
        if indep_test not in INDEP_NOT_MAXMARGIN.keys():
            raise ValueError("Test is not a valid independence test")
        if indep_test == "hsic" and compute_distkern == "euclidean":
            compute_distkern = "gaussian"
        self.indep_test_name = indep_test

        indep_kwargs = {
            "dcorr": {"bias": bias, "compute_distance": compute_distkern},
            "hsic": {"bias": bias, "compute_kernel": compute_distkern},
            "hhg": {"compute_distance": compute_distkern},
            "mgc": {"compute_distance": compute_distkern},
            "kmerf": {"forest_type": "classifier"},
            "rv": {},
            "cca": {},
        }
        self.indep_test = INDEP_NOT_MAXMARGIN[indep_test](
            **indep_kwargs[indep_test], **kwargs
        )
        self.is_fast = False
        self.bias = bias

        IndependenceTest.__init__(self, compute_distance=compute_distkern, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the Maximal Margin test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed Maximal Margin statistic.
        """
        stat = np.max(
            [
                self.indep_test.statistic(
                    x[:, i].reshape(-1, 1), y[:, j].reshape(-1, 1)
                )
                for i in range(x.shape[1])
                for j in range(y.shape[1])
            ]
        )
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the Maximal Margin test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Only applies to ``"Dcorr"`` and ``"Hsic"``.
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed Maximal Margin statistic.
        pvalue : float
            The computed Maximal Margin p-value.
        dict
            A dictionary containing optional parameters for tests that return them.
            See the relevant test in :mod:`hyppo.independence`.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import MaxMargin
        >>> x = np.arange(100)
        >>> y = x
        >>> stat, pvalue = MaxMargin("Dcorr").test(x, y)
        >>> '%.1f, %.3f' % (stat, pvalue)
        '1.0, 0.000'
        """
        check_input = _CheckInputs(
            x,
            y,
            reps=reps,
        )
        x, y = check_input()

        if (
            auto
            and x.shape[1] == 1
            and y.shape[1] == 1
            and self.compute_distance == "euclidean"
            and self.indep_test_name == "dcorr"
        ):
            self.is_fast = True

        if auto and x.shape[0] > 20 and self.indep_test_name in ["dcorr", "hsic"]:
            stat, pvalue = chi2_approx(self.statistic, x, y)
            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = None
        else:
            stat, pvalue = super(MaxMargin, self).test(
                x, y, reps, workers, is_distsim=False, random_state=None
            )

        return IndependenceTestOutput(stat, pvalue)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...tools import power, rot_ksamp
from .. import Hotelling


class TestHotelling:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue",
        [
            (1000, 0.00253015392620975, 0.997473047413008),
            (100, 8.24e-5, 0.953158769593477),
        ],
    )
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2, noise=False)
        stat, pvalue = Hotelling().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=1)


class TestHotellingTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "Hotelling",
            sim_type="ksamp",
            sim="multimodal_independence",
            k=2,
            n=100,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_less

from ...tools import nonlinear_process
from .. import MGCX


class TestMGCXStat:
    def test_multiple_lags(self):
        n = 10
        x = np.arange(1, n + 1).reshape(n, 1)
        y = np.arange(1, n + 1).reshape(n, 1)
        pvalue = MGCX(max_lag=3).test(x, y)[1]
        assert_almost_equal(pvalue, 1 / 1000, decimal=2)

    def test_nonlinear(self):
        np.random.seed(123456789)

        x, y = nonlinear_process(120, lag=1)
        _, pvalue, mgcx_dict = MGCX(max_lag=1).test(x, y)
        opt_lag = mgcx_dict["opt_lag"]

        assert_almost_equal(pvalue, 1 / 1000, decimal=2)
        assert_almost_equal(opt_lag, 1, decimal=0)

    # TO DO: Fix when repeated values are fixed in classic MGC.
    # def test_lag0(self):
    #     x, y = cross_corr_ar(20, lag=1, phi=0.9)

    #     stat1 = MGC().test(x, y)[0]
    #     stat2 = MGCX(max_lag=0).test(x, y)[0]

    #     assert_almost_equal(stat1, stat2, decimal=0)

    def test_optimal_scale_linear(self):
        n = 10
        x = np.arange(n)
        y = x
        mgcx_dict = MGCX().test(x, y, reps=100)[2]
        opt_scale = mgcx_dict["opt_scale"]
        assert_almost_equal(opt_scale[0], n, decimal=0)
        assert_almost_equal(opt_scale[1], n, decimal=0)

    def test_optimal_scale_nonlinear(self):
        n = 7
        x = np.arange(n)
        y = x**2
        mgcx_dict = MGCX().test(x, y, reps=100)[2]
        opt_scale = mgcx_dict["opt_scale"]
        assert_array_less(opt_scale[0], n)
        assert_array_less(opt_scale[1], n)

    def test_rep_linear(self):
        n = 10
        x = np.arange(n)
        y = x
        mgcx_dict = MGCX().test(x, y, reps=100, random_state=2)[2]
        opt_scale = mgcx_dict["opt_scale"]
        mgcx_dict2 = MGCX().test(x, y, reps=100, random_state=2)[2]
        opt_scale2 = mgcx_dict2["opt_scale"]

        assert opt_scale == opt_scale2

    def test_rep_nonlinear(self):
        n = 7
        x = np.arange(n)
        y = x**2
        mgcx_dict = MGCX().test(x, y, reps=100, random_state=2)[2]
        opt_scale = mgcx_dict["opt_scale"]
        mgcx_dict2 = MGCX().test(x, y, reps=100, random_state=2)[2]
        opt_scale2 = mgcx_dict2["opt_scale"]

        assert opt_scale == opt_scale2



import warnings
from typing import NamedTuple

import numpy as np
from scipy.stats import multiscale_graphcorr

from ..tools import compute_dist
from ._utils import _CheckInputs
from .base import IndependenceTest


class MGCTestOutput(NamedTuple):
    stat: float
    pvalue: float
    mgc_dict: dict


class MGC(IndependenceTest):
    r"""
    Multiscale Graph Correlation (MGC) test statistic and p-value.

    Specifically, for each point, MGC finds the :math:`k`-nearest neighbors for
    one property (e.g. cloud density), and the :math:`l`-nearest neighbors for
    the other property (e.g. grass wetness)
    :footcite:p:`vogelsteinDiscoveringDecipheringRelationships2019`.
    This pair :math:`(k, l)` is
    called the "scale". A priori, however, it is not know which scales will be
    most informative. So, MGC computes all distance pairs, and then efficiently
    computes the distance correlations for all scales. The local correlations
    illustrate which scales are relatively informative about the relationship.
    The key, therefore, to successfully discover and decipher relationships
    between disparate data modalities is to adaptively determine which scales
    are the most informative, and the geometric implication for the most
    informative scales. Doing so not only provides an estimate of whether the
    modalities are related, but also provides insight into how the
    determination was made. This is especially important in high-dimensional
    data, where simple visualizations do not reveal relationships to the
    unaided human eye. Characterizations of this implementation in particular
    have been derived from and benchmarked within
    :footcite:t:`pandaHyppoMultivariateHypothesis2021`.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.

    Notes
    -----
    A description of the process of MGC and applications on neuroscience data
    can be found in
    :footcite:t:`vogelsteinDiscoveringDecipheringRelationships2019`.
    It is performed using the following steps:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. Let :math:`D^x` be the :math:`n \times n`
    distance matrix of :math:`x` and :math:`D^y` be the :math:`n \times n` be
    the distance matrix of :math:`y`. :math:`D^x` and :math:`D^y` are
    modified to be mean zero columnwise. This results in two
    :math:`n \times n` distance matrices :math:`A` and :math:`B` (the
    centering and unbiased modification)
    :footcite:p:`shenDistanceCorrelationMultiscale2020`.

    + For all values :math:`k` and :math:`l` from :math:`1, ..., n`,

       * The :math:`k`-nearest neighbor and :math:`l`-nearest neighbor graphs
         are calculated for each property. Here, :math:`G_k (i, j)` indicates
         the :math:`k`-smallest values of the :math:`i`-th row of :math:`A`
         and :math:`H_l (i, j)` indicates the :math:`l` smallested values of
         the :math:`i`-th row of :math:`B`

       * The local
         correlations are summed and normalized using the following statistic:

         .. math::

            c^{kl} = \frac{\sum_{ij} A G_k B H_l}
                            {\sqrt{\sum_{ij} A^2 G_k \times \sum_{ij} B^2 H_l}}

    + The MGC test statistic is the smoothed optimal local correlation of
      :math:`\{ c^{kl} \}`. Denote the smoothing operation as :math:`R(\cdot)`
      (which essentially set all isolated large correlations) as 0 and
      connected large correlations the same as before, see
      :footcite:t:`shenDistanceCorrelationMultiscale2020`.)
      MGC is,

      .. math::

         \mathrm{MGC}_n (x, y) = \max_{(k, l)} R \left(c^{kl} \left( x_n, y_n \right)
                                                    \right)

    The test statistic returns a value between :math:`(-1, 1)` since it is
    normalized.

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", **kwargs):
        # set is_distance to true if compute_distance is None
        self.is_distance = False
        if not compute_distance:
            self.is_distance = True
        IndependenceTest.__init__(self, compute_distance=compute_distance, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the MGC test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed MGC statistic.
        """
        distx = x
        disty = y

        if not self.is_distance:
            distx, disty = compute_dist(
                x, y, metric=self.compute_distance, **self.kwargs
            )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            mgc = multiscale_graphcorr(distx, disty, compute_distance=None, reps=0)

        stat = mgc.stat
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the MGC test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed MGC statistic.
        pvalue : float
            The computed MGC p-value.
        mgc_dict : dict
            Contains additional useful returns containing the following keys:

                - mgc_map : ndarray of float
                    A 2D representation of the latent geometry of the relationship.
                - opt_scale : (int, int)
                    The estimated optimal scale as a ``(x, y)`` pair.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import MGC
        >>> x = np.arange(100)
        >>> y = x
        >>> stat, pvalue, _ = MGC().test(x, y)
        >>> '%.1f, %.3f' % (stat, pvalue)
        '1.0, 0.001'

        In addition, the inputs can be distance matrices. Using this is the,
        same as before, except the ``compute_distance`` parameter must be set
        to ``None``.

        >>> import numpy as np
        >>> from hyppo.independence import MGC
        >>> x = np.ones((10, 10)) - np.identity(10)
        >>> y = 2 * x
        >>> mgc = MGC(compute_distance=None)
        >>> stat, pvalue, _ = mgc.test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '0.0, 1.00'
        """
        check_input = _CheckInputs(
            x,
            y,
            reps=reps,
        )
        x, y = check_input()

        # add test for redundant rows
        if (
            np.unique(x, axis=0).shape[0] != x.shape[0]
            or np.unique(y, axis=0).shape[0] != y.shape[0]
        ):
            warnings.warn(
                "Input x has {} redundant rows, and input y has {} redundant "
                "rows. MGC Map will be of shape ({}, {}).".format(
                    x.shape[0] - np.unique(x, axis=0).shape[0],
                    y.shape[0] - np.unique(y, axis=0).shape[0],
                    np.unique(x, axis=0).shape[0],
                    np.unique(y, axis=0).shape[0],
                ),
                RuntimeWarning,
            )

        x, y = compute_dist(x, y, metric=self.compute_distance, **self.kwargs)
        self.is_distance = True

        # using our joblib implementation instead of multiprocessing backend in
        # scipy gives significantly faster results
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            _, _, mgc_dict = multiscale_graphcorr(x, y, compute_distance=None, reps=0)
        mgc_dict.pop("null_dist")

        stat, pvalue = super(MGC, self).test(
            x, y, reps, workers, random_state=random_state
        )
        self.mgc_dict = mgc_dict

        return MGCTestOutput(stat, pvalue, mgc_dict)



from ._utils import k_sample_transform
from .disco import DISCO
from .energy import Energy
from .friedman_rafsky import FriedmanRafsky
from .hotelling import Hotelling
from .ksamp import KSample
from .ksamplehhg import KSampleHHG
from .manova import MANOVA
from .mean_embedding import MeanEmbeddingTest, mean_embed_distance
from .mmd import MMD
from .smoothCF import SmoothCFTest, smooth_cf_distance

__all__ = [s for s in dir()]  # add imported tests to __all__


KSAMP_TESTS = {
    "disco": DISCO,
    "energy": Energy,
    "hotelling": Hotelling,
    "ksample": KSample,
    "manova": MANOVA,
    "mmd": MMD,
    "smoothCF": SmoothCFTest,
    "mean_embedding": MeanEmbeddingTest,
    "ksamplehhg": KSampleHHG,
    "friedman_rafsky": FriedmanRafsky,
}



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ...independence import Dcorr
from ...tools import cross_corr_ar, nonlinear_process
from .. import DcorrX


class TestDcorrXStat:
    def test_zero_var(self):
        x = np.ones(4)
        y = np.arange(4)
        stat = DcorrX().test(x, y)[0]
        assert_almost_equal(stat, 0.0)

    def test_multiple_lags(self):
        x = np.ones(6)
        y = np.arange(1, 7)
        stat = DcorrX(max_lag=3).test(x, y, reps=0)[0]
        assert_almost_equal(stat, 0.0)

    @pytest.mark.parametrize("n", [100, 200])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    @pytest.mark.parametrize("obs_opt_lag", [1])
    def test_correlated_oned(self, n, obs_pvalue, obs_opt_lag):
        np.random.seed(123456789)

        x, y = cross_corr_ar(n, lag=1, phi=0.9, sigma=0.1)
        _, pvalue, dcorrx_dict = DcorrX(max_lag=1).test(x, y)
        opt_lag = dcorrx_dict["opt_lag"]

        assert_almost_equal(pvalue, obs_pvalue, decimal=2)
        assert_almost_equal(opt_lag, obs_opt_lag, decimal=0)

    @pytest.mark.parametrize("n", [100, 200])
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    @pytest.mark.parametrize("obs_opt_lag", [1])
    def test_rep(self, n, obs_pvalue, obs_opt_lag):
        x, y = cross_corr_ar(n, lag=1, phi=0.9, sigma=0.1)
        _, pvalue, dcorrx_dict = DcorrX(max_lag=1).test(x, y, random_state=2)
        opt_lag = dcorrx_dict["opt_lag"]

        _, pvalue2, dcorr_dict2 = DcorrX(max_lag=1).test(x, y, random_state=2)
        opt_lag2 = dcorrx_dict["opt_lag"]

        assert pvalue == pvalue2
        assert opt_lag == opt_lag2

    def test_lag0(self):
        # Dependent
        x, y = nonlinear_process(100, lag=1)

        stat1 = Dcorr().test(x, y, auto=False, reps=0)[0]
        stat2 = DcorrX(max_lag=0).test(x, y, reps=0)[0]

        assert_almost_equal(stat1, stat2, decimal=2)

        # Independent
        x = np.array([[0.95476444], [0.36741182], [0.37473209], [-0.62469378]])
        y = np.array([[0.86515308], [-0.01985114], [0.21544295], [1.56443374]])
        stat = Dcorr().test(x, y, auto=False, reps=0)[0]
        statx = DcorrX(max_lag=0).test(x, y, reps=0)[0]

        assert_almost_equal(stat, statx, decimal=2)



import numpy as np
import pytest
from numpy.testing import (
    assert_almost_equal,
    assert_approx_equal,
    assert_equal,
    assert_warns,
)

from ...tools import linear, multimodal_independence, power, spiral
from .. import MGC


class TestMGCStat(object):
    """Test validity of MGC test statistic"""

    @pytest.mark.parametrize(
        "sim, obs_stat, obs_pvalue",
        [
            (linear, 0.97, 1 / 1000),  # test linear simulation
            (spiral, 0.163, 1 / 1000),  # test spiral simulation
        ],
    )
    def test_oned(self, sim, obs_stat, obs_pvalue):
        np.random.seed(12345678)

        # generate x and y
        x, y = sim(n=100, p=1)

        # test stat and pvalue
        stat1 = MGC().statistic(x, y)
        stat2, pvalue, _ = MGC().test(x, y)
        assert_approx_equal(stat1, obs_stat, significant=1)
        assert_approx_equal(stat2, obs_stat, significant=1)
        assert_approx_equal(pvalue, obs_pvalue, significant=1)

    @pytest.mark.parametrize(
        "sim, obs_stat, obs_pvalue",
        [
            (linear, 0.463, 1 / 1000),  # test linear simulation
            (spiral, 0.091, 0.01),  # test spiral simulation
        ],
    )
    def test_fived(self, sim, obs_stat, obs_pvalue):
        np.random.seed(12345678)

        # generate x and y
        x, y = sim(n=100, p=5)

        # test stat and pvalue
        stat1 = MGC().statistic(x, y)
        stat2, pvalue, _ = MGC().test(x, y)
        assert_approx_equal(stat1, obs_stat, significant=1)
        assert_approx_equal(stat2, obs_stat, significant=1)
        assert_approx_equal(pvalue, obs_pvalue, significant=1)

    @pytest.mark.parametrize(
        "sim, obs_stat, obs_pvalue",
        [
            (linear, 0.97, 1 / 1000),
        ],  # test linear simulation
    )
    def test_rep(self, sim, obs_stat, obs_pvalue):
        x, y = sim(n=100, p=5)
        stat1, pvalue1, _ = MGC().test(x, y, random_state=2)
        stat2, pvalue2, _ = MGC().test(x, y, random_state=2)

        assert stat1 == stat2
        assert pvalue1 == pvalue2


class TestMGCTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "MGC",
            sim_type="indep",
            sim="multimodal_independence",
            n=50,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)


class TestMGCErrorWarnings:
    def test_redundancy_warning(self):
        x = np.hstack((np.arange(0, 6), 5, 5, 5, 5, 5))
        y = np.hstack((np.arange(0, 6), 5, 5, 5, 5, 5))
        assert_warns(RuntimeWarning, MGC().test, x, y)



from abc import ABC, abstractmethod
from typing import NamedTuple

import numpy as np
from joblib import Parallel, delayed
from sklearn.utils import check_random_state

from ..tools import compute_dist
from sklearn.utils import check_random_state


class TimeSeriesTestOutput(NamedTuple):
    stat: float
    pvalue: float
    test_dict: dict


class TimeSeriesTest(ABC):
    """
    A base class for a time-series test.

    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        function.
    max_lag : float, default: 0
        The maximium lag to consider when computing the test statistics and p-values.
    **kwargs
        Arbitrary keyword arguments for ``compute_distance``.
    """

    def __init__(self, compute_distance=None, max_lag=0, **kwargs):
        # set statistic and p-value
        self.stat = None
        self.pvalue = None
        self.max_lag = max_lag
        self.kwargs = kwargs

        # set compute_distance kernel
        self.compute_distance = compute_distance

        super().__init__()

    @abstractmethod
    def statistic(self, x, y):
        """
        Calulates the time-series test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        """

    @abstractmethod
    def test(self, x, y, reps=1000, workers=1, random_state=None, is_distsim=False):
        """
        Calulates the time-series test test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be distance matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        is_distsim : bool, default: False
            Whether or not ``x`` and ``y`` are input matrices.

        Returns
        -------
        stat : float
            The computed time-series independence test statistic.
        pvalue : float
            The computed time-series independence p-value.
        null_dist : list
            The time-series independence p-value.
        """
        # calculate observed test statistic
        stat_list = self.statistic(x, y)
        stat = stat_list[0]

        # make RandomState seeded array
        if random_state is not None:
            rng = check_random_state(random_state)
            random_state = rng.randint(np.iinfo(np.int32).max, size=reps)

        # make random array
        else:
            random_state = np.random.randint(np.iinfo(np.int32).max, size=reps)

        # calculate null distribution
        null_dist = np.array(
            Parallel(n_jobs=workers)(
                [
                    delayed(_perm_stat)(self.statistic, x, y, rng, is_distsim)
                    for rng in random_state
                ]
            )
        )
        pvalue = (1 + (null_dist >= stat).sum()) / (1 + reps)
        self.pvalue = pvalue
        self.null_dist = null_dist

        return stat, pvalue, stat_list


def _perm_stat(calc_stat, distx, disty, random_state, is_distsim):
    """Permutes the test statistics."""
    n = distx.shape[0]
    block_size = int(np.ceil(np.sqrt(n)))
    rng = check_random_state(random_state)
    perm_index = np.r_[
        [np.arange(t, t + block_size) for t in rng.choice(n, n // block_size + 1)]
    ].flatten()[:n]
    perm_index = np.mod(perm_index, n)
    if is_distsim:
        permy = disty[np.ix_(perm_index, perm_index)]
    else:
        permy = disty[perm_index]

    # calculate permuted statics, store in null distribution
    perm_stat = calc_stat(distx, permy)[0]

    return perm_stat



import numpy as np
from numba import jit
from hyppo.ksample.base import KSampleTest
from hyppo.ksample._utils import _CheckInputs
from sklearn.metrics import pairwise_distances
from scipy.stats import ks_2samp


class KSampleHHG(KSampleTest):
    r"""
    HHG 2-Sample test statistic.
    
    This is a 2-sample multivariate test based on univariate test statistics.
    It inherits the computational complexity from the unvariate tests to achieve
    faster speeds than classic multivariate tests.
    The univariate test used is the Kolmogorov-Smirnov 2-sample test.
    :footcite:p:`hellerMultivariateTestsOfAssociation2016`.
    
    Parameters
    ----------
    compute_distance : str, callable, or None, default: "euclidean"
        A function that computes the distance among the samples within each
        data matrix.
        Valid strings for ``compute_distance`` are, as defined in
        :func:`sklearn.metrics.pairwise_distances`,

            - From scikit-learn: [``"euclidean"``, ``"cityblock"``, ``"cosine"``,
              ``"l1"``, ``"l2"``, ``"manhattan"``] See the documentation for
              :mod:`scipy.spatial.distance` for details
              on these metrics.
            - From scipy.spatial.distance: [``"braycurtis"``, ``"canberra"``,
              ``"chebyshev"``, ``"correlation"``, ``"dice"``, ``"hamming"``,
              ``"jaccard"``, ``"kulsinski"``, ``"mahalanobis"``, ``"minkowski"``,
              ``"rogerstanimoto"``, ``"russellrao"``, ``"seuclidean"``,
              ``"sokalmichener"``, ``"sokalsneath"``, ``"sqeuclidean"``,
              ``"yule"``] See the documentation for :mod:`scipy.spatial.distance` for
              details on these metrics.

        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already distance
        matrices. To call a custom function, either create the distance matrix
        before-hand or create a function of the form ``metric(x, **kwargs)``
        where ``x`` is the data matrix for which pairwise distances are
        calculated and ``**kwargs`` are extra arguements to send to your custom
        
     **kwargs
         Arbitrary keyword arguments for ``compute_distance``.
    
    Notes
    -----
    The statistic can be derived as follows:
    :footcite:p:`hellerMultivariateTestsOfAssociation2016`.
    
    Let :math:`x`, :math:`y` be :math:`(n, p)`, :math:`(m, p)` samples of random variables
    :math:`X` and :math:`Y \in \R^p` . Let there be a center point
    :math:`\in \R^p`. 
    For every sample :math:`i`, calculate the distances from the center point
    in :math:`x` and :math:`y` and denote this as :math:`d_x(x_i)`
    and :math:`d_y(y_i)`. This will create a 1D collection of distances for each
    sample group.
    
    Then apply the KS 2-sample test on these center-point distances. This classic test
    compares the empirical distribution function of the two samples and takes
    the supremum of the difference between them. See Notes under scipy.stats.ks_2samp
    for more details.
    
    To achieve better power, the above process is repeated with each sample point 
    :math:`x_i` and :math:`y_i` as center points. The resultant :math:`n+m` p-values
    are then pooled for use in the Bonferroni test of the global null hypothesis.
    The HHG statistic is the KS stat associated with the smallest p-value from the pool,
    while the HHG p-value is the smallest p-value multipled by the number of sample points.
    
    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_distance="euclidean", **kwargs):
        self.compute_distance = compute_distance
        KSampleTest.__init__(self, compute_distance=compute_distance, **kwargs)

    def statistic(self, x, y):
        """
        Calculates K-Sample HHG test statistic.
        
        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` and  are the number of samples and `p` is the number of
            dimensions.
            
        Returns
        -------
        stat : float
            The computed KS test statistic associated with the lowest p-value.
        """
        xy = np.concatenate((x, y), axis=0)
        distxy = _centerpoint_dist(xy, self.compute_distance, 1)
        distx = distxy[:, 0 : len(x)]
        disty = distxy[:, len(x) : len(x) + len(y)]
        stats, pvalues = _distance_score(distx, disty)
        minP = min(pvalues)
        stat = stats[pvalues.index(minP)]
        self.minP = minP
        self.stat = stat
        return self.stat

    def test(self, x, y):
        """
        Calculates K-Sample HHG test statistic and p-value.
        
        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` and `m` are the number of samples and `p` is the number of
            dimensions.
            
        Returns
        -------
        stat : float
            The computed KS test statistic associated with the lowest p-value.
        pvalue : float
            The computed HHG pvalue. Equivalent to the lowest p-value multiplied by the total number
            of samples.
        """
        check_input = _CheckInputs(inputs=[x, y],)
        x, y = check_input()
        N = x.shape[0] + y.shape[0]

        stat = self.statistic(x, y)
        pvalue = self.minP * N
        return stat, pvalue


def _centerpoint_dist(xy, metric, workers=1, **kwargs):
    """Gives pairwise distances - each row corresponds to center-point distances
    where one sample point is the center point"""
    distxy = pairwise_distances(xy, metric=metric, n_jobs=workers, **kwargs)
    return distxy


def _distance_score(distx, disty):
    dist1, dist2 = _group_distances(distx, disty)
    stats = []
    pvalues = []
    for i in range(len(distx)):
        stat, pvalue = ks_2samp(dist1[i], dist2[i])
        stats.append(stat)
        pvalues.append(pvalue)
    return stats, pvalues


@jit(nopython=True, cache=True)
def _group_distances(distx, disty):  # pragma: no cover
    dist1 = []
    dist2 = []
    for i in range(len(distx)):
        distancex = np.delete(distx[i], 0)
        distancey = np.delete(disty[i], 0)
        distancex = distancex.reshape(-1)
        distancey = distancey.reshape(-1)
        dist1.append(distancex)
        dist2.append(distancey)
    return dist1, dist2



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises

from ...tools import power, rot_ksamp
from .. import KSample


class TestKSample:
    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue, indep_test",
        [(1000, 4.28e-7, 1.0, "CCA"), (100, 0.045646974150778084, 0.001, "Dcorr")],
    )
    def test_twosamp_linear_oned(self, n, obs_stat, obs_pvalue, indep_test):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", n, 1, k=2)
        stat, pvalue = KSample(indep_test).test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=1)
        assert_almost_equal(pvalue, obs_pvalue, decimal=1)

    def test_rf(self):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", 50, 1, k=2)
        stat, _, _ = KSample("KMERF").test(x, y, reps=0)

        assert_almost_equal(stat, 0.27140550503906097, decimal=1)

    def test_maxmargin(self):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", 50, 1, k=2)
        stat, _ = KSample(["MaxMargin", "Dcorr"]).test(x, y, reps=0)

        assert_almost_equal(stat, 0.0317, decimal=1)

    @pytest.mark.parametrize(
        "n, obs_stat, obs_pvalue, indep_test",
        [(100, 8.24e-5, 0.001, "Dcorr")],
    )
    def test_rep(self, n, obs_stat, obs_pvalue, indep_test):
        x, y = rot_ksamp("linear", n, 1, k=2)
        stat, pvalue = KSample(indep_test).test(x, y)
        stat2, pvalue2 = KSample(indep_test).test(x, y)

        assert stat == stat2
        assert pvalue == pvalue2


class TestKSampleErrorWarn:
    """Tests errors and warnings derived from MGC."""

    def test_no_indeptest(self):
        # raises error if not indep test
        assert_raises(ValueError, KSample, "abcd")

    def test_no_second_maxmargin(self):
        np.random.seed(123456789)
        x, y = rot_ksamp("linear", 50, 1, k=2)
        assert_raises(ValueError, KSample, ["MaxMargin", "abcd"])


class TestKSampleTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "CCA",
            sim_type="ksamp",
            sim="multimodal_independence",
            k=2,
            n=100,
            p=1,
            alpha=0.05,
            auto=True,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np

from .indep_sim import SIMULATIONS


def _2samp_rotate(sim, x, y, p, degree=90, pow_type="samp"):
    """Generate an independence simulation, rotate it to produce another."""
    angle = np.radians(degree)
    data = np.hstack([x, y])
    same_shape = [
        "joint_normal",
        "logarithmic",
        "sin_four_pi",
        "sin_sixteen_pi",
        "two_parabolas",
        "square",
        "diamond",
        "circle",
        "ellipse",
        "multiplicative_noise",
        "multimodal_independence",
    ]
    if sim in same_shape:
        rot_shape = 2 * p
    else:
        rot_shape = p + 1
    rot_mat = np.identity(rot_shape)
    if pow_type == "dim":
        if sim not in [
            "exponential",
            "cubic",
            "spiral",
            "uncorrelated_bernoulli",
            "fourth_root",
        ]:
            for i in range(rot_shape):
                mat = np.random.normal(size=(rot_shape, 1))
                mat = mat / np.sqrt(np.sum(mat**2))
                if i == 0:
                    rot = mat
                else:
                    rot = np.hstack([rot, mat])
                rot_mat, _ = np.linalg.qr(rot)
                if (p % 2) == 1:
                    rot_mat[0] *= -1
        else:
            rot_mat[np.ix_((0, -1), (0, -1))] = np.array(
                [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
            )
    elif pow_type == "samp":
        rot_mat[np.ix_((0, 1), (0, 1))] = np.array(
            [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
        )
    else:
        raise ValueError("pow_type not a valid flag ('dim', 'samp')")
    rot_data = (rot_mat @ data.T).T

    if sim in same_shape:
        x_rot, y_rot = np.hsplit(rot_data, 2)
    else:
        x_rot, y_rot = np.hsplit(rot_data, [-p])

    return x_rot, y_rot


def rot_ksamp(sim, n, p, k=2, noise=True, degree=90, pow_type="samp", **kwargs):
    r"""
    Rotates input simulations to produce a `k`-sample simulation.

    Parameters
    ----------
    sim : str
        The name of the simulation (from the :mod:`hyppo.tools` module) that is to be
        rotated.
    n : int
        The number of samples desired by ``sim`` (>= 5).
    p : int
        The number of dimensions desired by ``sim`` (>= 1).
    k : int, default: 2
        The number of groups to simulate.
    noise : bool, default: True
        Whether or not to include noise in the simulation.
    degree : float or list of float, default: 90
        The number of degrees to rotate the input simulation by (in first dimension).
        The list must be the same size as ``k - 1``.
    pow_type : "samp", "dim", default: "samp"
        Simulation type, (increasing sample size or dimension).
    **kwargs
        Additional keyword arguments for the independence simulation.

    Returns
    -------
    sims : list of ndarray
        Rotated data matrices. ``sims`` is a list of arrays of shape ``(n, p+1)``
        or ``(n, 2p)`` depending on the independence simulation. Here, `n`
        is the number of samples and `p` is the number of dimensions.
    """
    if sim not in SIMULATIONS.keys():
        raise ValueError(
            "Not valid simulation, must be one of {}".format(SIMULATIONS.keys())
        )

    if (k - 1) > 1:
        if type(degree) is list:
            if (k - 1) != len(degree):
                raise ValueError(
                    "k={}, so length of degree must be {}, got {}".format(
                        k, k - 1, len(degree)
                    )
                )
        else:
            if (k - 1) != 1:
                raise ValueError(
                    "k={}, so degree must be type int or float, got {}".format(
                        k, type(degree)
                    )
                )

    if sim == "multimodal_independence":
        sims = [np.hstack(SIMULATIONS[sim](n, p, **kwargs)) for _ in range(k)]
    else:
        if sim != "multiplicative_noise":
            kwargs["noise"] = noise
        x, y = SIMULATIONS[sim](n, p, **kwargs)
        if (k - 1) == 1:
            sims = [
                np.hstack([x, y]),
                np.hstack(
                    _2samp_rotate(sim, x, y, p, degree=degree, pow_type=pow_type)
                ),
            ]
        else:
            sims = [np.hstack([x, y])] + [
                np.hstack(_2samp_rotate(sim, x, y, p, degree=deg, pow_type=pow_type))
                for deg in degree
            ]

    return sims


def gaussian_3samp(n, epsilon=1, weight=0, case=1):
    r"""
    Generates 3 sample of gaussians corresponding to 5 cases.

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 5).
    epsilon : float, default: 1
        The amount to translate simulation by (amount  depends on case).
    weight : float, default: False
        Number between 0 and 1 corresponding to weight of the second Gaussian
        (used in case 4 and 5 to produce a mixture of Gaussians).
    case : 1, 2, 3, 4, 5, default: 1
        The case in which to evaluate statistical power for each test.

    Returns
    -------
    sims : list of ndarray
        List of 3 2-dimensional multivariate Gaussian each
        corresponding to the desired case.
    """
    if n < 5:
        raise ValueError("n must be >= 5, got {}".format(n))

    old_case = case
    if old_case == 4:
        case = 2
    elif old_case == 5:
        case = 3
    sigma = np.identity(2)
    mu1 = [0] * 3
    mu2 = [0] * 3

    if case == 1:
        pass
    elif case == 2:
        mu2 = [0, 0, epsilon]
    elif case == 3:
        mu1 = [0, -epsilon / 2, epsilon / 2]
        mu2 = [
            (np.sqrt(3) / 3) * epsilon,
            -(np.sqrt(3) / 6) * epsilon,
            -(np.sqrt(3) / 6) * epsilon,
        ]
    else:
        raise ValueError("Not valid case, must be 1, 2, or 3")

    means = list(zip(mu1, mu2))
    sims = [np.random.multivariate_normal(mean, sigma, n) for mean in means]
    if old_case == 4:
        sims[-1] = (1 - weight) * sims[-1] + weight * np.random.multivariate_normal(
            means[-1], sigma * 1.5, n
        )
    elif old_case == 5:
        sims = [
            (1 - weight) * sims[i]
            + weight * np.random.multivariate_normal(means[i], sigma * 1.5, n)
            for i in range(len(sims))
        ]

    return sims


KSAMP_SIMS = {
    "rot_ksamp": rot_ksamp,
    "gaussian_3samp": gaussian_3samp,
}


def ksamp_sim(ksim, n, **kwargs):
    r"""
    `K`-sample simulation generator.

    Takes a simulation and the required parameters, and outputs the simulated
    data matrices.

    Parameters
    ----------
    sim : str
        The name of the simulation (from the :mod:`hyppo.tools module).
    n : int
        The number of samples desired by the simulation (>= 3).
    **kwargs
        Additional keyword arguements for the desired simulation.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices.
    """
    if ksim not in KSAMP_SIMS.keys():
        raise ValueError(
            "sim must be one of the following: {}".format(list(KSAMP_SIMS.keys()))
        )
    else:
        ksim = KSAMP_SIMS[ksim]

    return ksim(n=n, **kwargs)



import numpy as np

from ..tools import chi2_approx, compute_kern
from ._utils import _CheckInputs
from .base import IndependenceTest, IndependenceTestOutput
from .dcorr import Dcorr


class Hsic(IndependenceTest):
    r"""
    Hilbert Schmidt Independence Criterion (Hsic) test statistic and p-value.

    Hsic is a kernel based independence test and is a way to measure
    multivariate nonlinear associations given a specified kernel
    :footcite:p:`grettonKernelStatisticalTest2007`.
    The default choice is the Gaussian kernel, which uses the median distance
    as the bandwidth, which is a characteristic kernel that guarantees that
    Hsic is a consistent test
    :footcite:p:`grettonKernelStatisticalTest2007`
    :footcite:p:`grettonConsistentNonparametricTests2010`.

    Parameters
    ----------
    compute_kernel : str, callable, or None, default: "gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguements to send to your custom function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``compute_kernel``.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`grettonKernelStatisticalTest2007`:

    Hsic is closely related distance correlation (Dcorr), implemented in
    :class:`hyppo.independence.Dcorr`, and exchanges distance matrices
    :math:`D^x` and :math:`D^y` for kernel similarity matrices :math:`K^x` and
    :math:`K^y`. That is, let :math:`x` and :math:`y` be :math:`(n, p)` samples
    of random variables
    :math:`X` and :math:`Y`. Let :math:`K^x` be the :math:`n \times n`
    kernel similarity matrix of :math:`x` and :math:`K^y` be the :math:`n \times n` be
    the kernel similarity matrix of :math:`y`. The Hsic statistic is,

    .. math::

        \mathrm{Hsic}^b_n (x, y) = \frac{1}{n^2} \mathrm{tr} (D^x H D^y H)

    Hsic and Dcov are exactly equivalent in the sense that every valid kernel has a
    corresponding
    valid semimetric to ensure their equivalence, and vice versa
    :footcite:p:`shenExactEquivalenceDistance2020`
    :footcite:p:`sejdinovicEquivalenceDistancebasedRKHSbased2013`.
    In other words, every Dcorr test is also an Hsic and vice versa. Nonetheless,
    implementations of Dcorr and Hsic use different metrics by default:
    Dcorr uses a Euclidean distance while Hsic uses a Gaussian median kernel.
    We consider the normalized version (see :class:`hyppo.independence`) for the
    transformation.

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`. The fast version of the test uses
    :meth:`hyppo.tools.chi2_approx`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_kernel="gaussian", bias=False, **kwargs):
        # set statistic and p-value
        self.compute_kernel = compute_kernel

        self.is_kernel = False
        if not compute_kernel:
            self.is_kernel = True
        self.bias = bias

        IndependenceTest.__init__(self, compute_distance=None, **kwargs)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the Hsic test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be kernel similarity
            matrices,
            where the shapes must both be ``(n, n)``.

        Returns
        -------
        stat : float
            The computed Hsic statistic.
        """
        distx = x
        disty = y

        if not self.is_kernel:
            kernx, kerny = compute_kern(x, y, metric=self.compute_kernel, **self.kwargs)
            distx = 1 - kernx / np.max(kernx)
            disty = 1 - kerny / np.max(kerny)

        # Hsic and Dcorr are equivalent, cannot use dcov otherwise fast is invalid
        stat = Dcorr(bias=self.bias, compute_distance=None).statistic(distx, disty)
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the Hsic test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples. That is, the shapes must be ``(n, p)`` and ``(n, q)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions. Alternatively, ``x`` and ``y`` can be kernel similarity
            matrices,
            where the shapes must both be ``(n, n)``.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed Hsic statistic.
        pvalue : float
            The computed Hsic p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import Hsic
        >>> x = np.arange(100)
        >>> y = x
        >>> stat, pvalue = Hsic().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '1.0, 0.00'
        """
        check_input = _CheckInputs(
            x,
            y,
            reps=reps,
        )
        x, y = check_input()

        if auto and x.shape[0] > 20:
            stat, pvalue = chi2_approx(self.statistic, x, y)
            self.stat = stat
            self.pvalue = pvalue
            self.null_dist = None
        else:
            x, y = compute_kern(x, y, metric=self.compute_kernel, **self.kwargs)
            self.is_kernel = True
            stat, pvalue = super(Hsic, self).test(
                x, y, reps, workers, random_state=None
            )

        return IndependenceTestOutput(stat, pvalue)



import numpy as np

from ..tools import check_ndarray_xy, check_reps, contains_nan, convert_xy_float64


class _CheckInputs:
    """Checks inputs for all independence tests"""

    def __init__(self, x, y, reps=None):
        self.x = x
        self.y = y
        self.reps = reps

    def __call__(self):
        check_ndarray_xy(self.x, self.y)
        contains_nan(self.x)
        contains_nan(self.y)
        self.x, self.y = self.check_dim_xy()
        self.x, self.y = convert_xy_float64(self.x, self.y)
        self._check_min_samples()
        self._check_variance()

        if self.reps:
            check_reps(self.reps)

        return self.x, self.y

    def check_dim_xy(self):
        """Convert x and y to proper dimensions"""
        if self.x.ndim == 1:
            self.x = self.x[:, np.newaxis]
        elif self.x.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `x`, found shape " "{}".format(self.x.shape)
            )
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
        elif self.y.ndim != 2:
            raise ValueError(
                "Expected a 2-D array `y`, found shape " "{}".format(self.y.shape)
            )

        self._check_nd_indeptest()

        return self.x, self.y

    def _check_nd_indeptest(self):
        """Check if number of samples is the same"""
        nx, _ = self.x.shape
        ny, _ = self.y.shape
        if nx != ny:
            raise ValueError(
                "Shape mismatch, x and y must have shape [n, p] and [n, q], found shape {} and {}".format(
                    self.x.shape, self.y.shape
                )
            )

    def _check_min_samples(self):
        """Check if the number of samples is at least 3"""
        nx = self.x.shape[0]
        ny = self.y.shape[0]

        if nx <= 3 or ny <= 3:
            raise ValueError("Number of samples is too low")

    def _check_variance(self):
        if np.var(self.x) == 0 or np.var(self.y) == 0:
            raise ValueError("Test cannot be run, one of the inputs has 0 variance")


def sim_matrix(model, x):
    """
    Computes the similarity matrix from a random forest.

    The model used must follow the scikit-learn API. That is, the model is a random
    forest class and is already trained (use :func:`fit`) before running this function.
    Also, :func:`apply` must be used to push input data down the trained forest. See
    :class:`sklearn.ensemble.RandomForestClassifier` and
    :class:`sklearn.ensemble.RandomForestRegressor` for similar classes.

    Parameters
    ----------
    model : class
        A trained random forest object for which the similarity matrix will be
        calculated.
    x : ndarray
        Input data matrice. ``x`` must have shape must be ``(n, p)`` where
        `n` is the number of samples and `p` are the number of
        dimensions.

    Returns
    -------
    proxMat : ndarray
        The proximity matrix induced by random kernel.
    """
    terminals = model.apply(x)
    ntrees = terminals.shape[1]

    prox_mat = sum(
        np.equal.outer(terminals[:, i], terminals[:, i]) for i in range(ntrees)
    )
    prox_mat = prox_mat / ntrees

    return prox_mat



import numpy as np

from ..tools import contains_nan


class _CheckInputs:
    def __init__(self, inputs, indep_test=None, reps=None):
        self.inputs = inputs
        self.reps = reps
        self.indep_test = indep_test

    def __call__(self):
        self._check_ndarray_inputs()
        for i in self.inputs:
            contains_nan(i)
        self.inputs = self.check_dim()
        self.inputs = self._convert_inputs_float64()
        self._check_indep_test()
        self._check_min_samples()

        return self.inputs

    def _check_ndarray_inputs(self):
        if len(self.inputs) < 2:
            raise ValueError("there must be at least 2 inputs")
        for i in self.inputs:
            if not isinstance(i, np.ndarray):
                raise ValueError("x and y must be ndarrays")

    def check_dim(self):
        # check if inputs are ndarrays
        new_inputs = []
        dims = []
        for i in self.inputs:
            # convert arrays of type (n,) to (n, 1)
            if i.ndim == 1:
                i = i[:, np.newaxis]
            elif i.ndim != 2:
                raise ValueError(
                    "Expected a 2-D array `i`, found shape " "{}".format(i.shape)
                )
            dims.append(i.shape[1])
            new_inputs.append(i)

        self._check_nd_ksampletest(dims)

        return new_inputs

    def _check_nd_ksampletest(self, dims):
        if len(set(dims)) > 1:
            raise ValueError(
                "Shape mismatch, inputs must have shape "
                "[n, p] and [m, p]. Found shapes {}".format(list(set(dims)))
            )

    def _convert_inputs_float64(self):
        return [np.asarray(i).astype(np.float64) for i in self.inputs]

    def _check_indep_test(self):
        tests = ["cca", "dcorr", "hhg", "rv", "hsic", "mgc", "kmerf"]
        if self.indep_test not in tests and self.indep_test is not None:
            raise ValueError("indep_test must be in {}".format(tests))

    def _check_min_samples(self):
        for i in self.inputs:
            if i.shape[0] <= 3:
                raise ValueError("Number of samples is too low")


def k_sample_transform(inputs, test_type="normal"):
    """
    Computes a `k`-sample transform of the inputs.

    For :math:`k` groups, this creates two matrices, the first vertically stacks the
    inputs.
    In order to use this function, the inputs must have the same number of dimensions
    :math:`p` and can have varying number of samples :math:`n`. The second output is a
    label
    matrix the one-hoc encodes the groups. The outputs are thus ``(N, p)`` and
    ``(N, k)`` where `N` is the total number of samples. In the case where the test
    a random forest based tests, it creates a ``(N, 1)`` where the entries are
    varlues from 1 to :math:`k` based on the number of samples.

    Parameters
    ----------
    inputs : list of ndarray
        A list of the inputs. All inputs must be ``(n, p)`` where `n` is the number
        of samples and `p` is the number of dimensions. `n` can vary between samples,
        but `p` must be the same among all the samples.
    test_type : {"normal", "rf"}, default: "normal"
        Whether to one-hoc encode the inputs ("normal") or use a one-dimensional
        categorical encoding ("rf").

    Returns
    -------
    u : ndarray
        The matrix of concatenated inputs of shape ``(N, p)``.
    v : ndarray
        The label matrix of shape ``(N, k)`` ("normal") or ``(N, 1)`` ("rf").
    """
    n_inputs = len(inputs)
    u = np.vstack(inputs)
    if np.var(u) == 0:
        raise ValueError("Test cannot be run, the inputs have 0 variance")

    if test_type == "rf":
        v = np.vstack(
            [np.repeat(i, inputs[i].shape[0]).reshape(-1, 1) for i in range(n_inputs)]
        )
    elif test_type == "normal":
        if n_inputs == 2:
            n1 = inputs[0].shape[0]
            n2 = inputs[1].shape[0]
            v = np.vstack([np.zeros((n1, 1)), np.ones((n2, 1))])
        else:
            vs = []
            for i in range(n_inputs):
                n = inputs[i].shape[0]
                encode = np.zeros(shape=(n, n_inputs))
                encode[:, i] = np.ones(shape=n)
                vs.append(encode)
            v = np.concatenate(vs)
    else:
        raise ValueError("test_type must be normal or rf")

    return u, v
